import re
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Sequence, Union, cast

from lilya._internal._path import clean_path
from lilya._utils import is_class_and_subclass
from lilya.middleware import DefineMiddleware
from lilya.routing import Path as LilyaPath, WebSocketPath as LilyaWebSocketPath, compile_path
from lilya.types import Receive, Scope, Send
from typing_extensions import Annotated, Doc

from esmerald.routing.apis.base import View
from esmerald.routing.base import Dispatcher
from esmerald.typing import Void, VoidType
from esmerald.utils.helpers import clean_string

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.interceptors.types import Interceptor
    from esmerald.openapi.schemas.v3_1_0.security_scheme import SecurityScheme
    from esmerald.permissions.types import Permission
    from esmerald.routing.router import HTTPHandler, WebhookHandler, WebSocketHandler
    from esmerald.types import Dependencies, ExceptionHandlerMap, Middleware, ParentType


class BaseMiddleware:
    def handle_middleware(
        self, handler: Any, base_middleware: List["Middleware"]
    ) -> List["Middleware"]:
        """
        Handles both types of middlewares for Gateway and WebSocketGateway
        """
        _middleware: List["Middleware"] = []

        if not is_class_and_subclass(handler, View) and not isinstance(handler, View):
            base_middleware += handler.middleware or []

        for middleware in base_middleware or []:
            if isinstance(middleware, DefineMiddleware):
                _middleware.append(middleware)
            else:
                _middleware.append(DefineMiddleware(middleware))  # type: ignore
        return _middleware


class GatewayUtil:

    def is_class_based(
        self, handler: Union["HTTPHandler", "WebSocketHandler", "ParentType"]
    ) -> bool:
        return bool(is_class_and_subclass(handler, View) or isinstance(handler, View))

    def is_handler(self, handler: Union["HTTPHandler", "WebSocketHandler", "ParentType"]) -> bool:
        return bool(not is_class_and_subclass(handler, View) and not isinstance(handler, View))

    def generate_operation_id(
        self, name: Union[str, None], handler: Union["HTTPHandler", "WebSocketHandler", View]
    ) -> str:
        """
        Generates an unique operation if for the handler.

        We need to be able to handle with edge cases when a view does not default a path like `/format` and a default name needs to be passed when its a class based view.
        """
        if self.is_class_based(handler.parent):
            operation_id = (
                handler.parent.__class__.__name__.lower() + f"_{name}" + handler.path_format
            )
        else:
            operation_id = name + handler.path_format

        operation_id = re.sub(r"\W", "_", operation_id)
        methods = list(handler.methods)  # type: ignore

        assert handler.methods  # type: ignore
        operation_id = f"{operation_id}_{methods[0].lower()}"
        return operation_id


class Gateway(LilyaPath, Dispatcher, BaseMiddleware, GatewayUtil):
    """
    `Gateway` object class used by Esmerald routes.

    The Gateway act as a brigde between the router handlers and
    the main Esmerald routing system.

    Read more about [Gateway](https://esmerald.dev/routing/routes/#gateway) and
    how to use it.

    **Example**

    ```python
    from esmerald import Esmerald. get

    @get()
    async def home() -> str:
        return "Hello, World!"

    Gateway(path="/home", handler=home)
    ```
    """

    __slots__ = (
        "_interceptors",
        "path",
        "handler",
        "name",
        "include_in_schema",
        "parent",
        "dependencies",
        "middleware",
        "exception_handlers",
        "interceptors",
        "permissions",
        "deprecated",
        "tags",
        "operation_id",
    )

    def __init__(
        self,
        path: Annotated[
            Optional[str],
            Doc(
                """
                Relative path of the `Gateway`.
                The path can contain parameters in a dictionary like format
                and if the path is not provided, it will default to `/`.

                **Example**

                ```python
                Gateway()
                ```

                **Example with parameters**

                ```python
                Gateway(path="/{age: int}")
                ```
                """
            ),
        ] = None,
        *,
        handler: Annotated[
            Union["HTTPHandler", View],
            Doc(
                """
            An instance of [handler](https://esmerald.dev/routing/handlers/#http-handlers).
            """
            ),
        ],
        name: Annotated[
            Optional[str],
            Doc(
                """
                The name for the Gateway. The name can be reversed by `path_for()`.
                """
            ),
        ] = None,
        include_in_schema: Annotated[
            bool,
            Doc(
                """
                Boolean flag indicating if it should be added to the OpenAPI docs.
                """
            ),
        ] = True,
        parent: Annotated[
            Optional["ParentType"],
            Doc(
                """
                Who owns the Gateway. If not specified, the application automatically it assign it.

                This is directly related with the [application levels](https://esmerald.dev/application/levels/).
                """
            ),
        ] = None,
        dependencies: Annotated[
            Optional["Dependencies"],
            Doc(
                """
                A dictionary of string and [Inject](https://esmerald.dev/dependencies/) instances enable application level dependency injection.
                """
            ),
        ] = None,
        middleware: Annotated[
            Optional[List["Middleware"]],
            Doc(
                """
                A list of middleware to run for every request. The middlewares of a Gateway will be checked from top-down or [Lilya Middleware](https://www.lilya.dev/middleware/) as they are both converted internally. Read more about [Python Protocols](https://peps.python.org/pep-0544/).
                """
            ),
        ] = None,
        interceptors: Annotated[
            Optional[Sequence["Interceptor"]],
            Doc(
                """
                A list of [interceptors](https://esmerald.dev/interceptors/) to serve the application incoming requests (HTTP and Websockets).
                """
            ),
        ] = None,
        permissions: Annotated[
            Optional[Sequence["Permission"]],
            Doc(
                """
                A list of [permissions](https://esmerald.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
            ),
        ] = None,
        exception_handlers: Annotated[
            Optional["ExceptionHandlerMap"],
            Doc(
                """
                A dictionary of [exception types](https://esmerald.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
            ),
        ] = None,
        deprecated: Annotated[
            Optional[bool],
            Doc(
                """
                Boolean flag for indicating the deprecation of the Gateway and to display it
                in the OpenAPI documentation..
                """
            ),
        ] = None,
        is_from_router: Annotated[
            bool,
            Doc(
                """
                Used by the `.add_router()` function of the `Esmerald` class indicating if the
                Gateway is coming from a router.
                """
            ),
        ] = False,
        security: Annotated[
            Optional[Sequence["SecurityScheme"]],
            Doc(
                """
                Used by OpenAPI definition, the security must be compliant with the norms.
                Esmerald offers some out of the box solutions where this is implemented.

                The [Esmerald security](https://esmerald.dev/openapi/) is available to automatically used.

                The security can be applied also on a [level basis](https://esmerald.dev/application/levels/).

                For custom security objects, you **must** subclass
                `esmerald.openapi.security.base.HTTPBase` object.
                """
            ),
        ] = None,
        tags: Annotated[
            Optional[Sequence[str]],
            Doc(
                """
                A list of strings tags to be applied to the *path operation*.

                It will be added to the generated OpenAPI documentation.

                **Note** almost everything in Esmerald can be done in [levels](https://esmerald.dev/application/levels/), which means
                these tags on a Esmerald instance, means it will be added to every route even
                if those routes also contain tags.
                """
            ),
        ] = None,
        operation_id: Annotated[
            Optional[str],
            Doc(
                """
                Unique operation id that allows distinguishing the same handler in different Gateways.

                Used for OpenAPI purposes.
                """
            ),
        ] = None,
    ) -> None:
        if not path:
            path = "/"
        if is_class_and_subclass(handler, View):
            handler = handler(parent=self)  # type: ignore

        if not is_from_router:
            self.path = clean_path(path + handler.path)
        else:
            self.path = clean_path(path)

        self.methods = getattr(handler, "http_methods", None)

        if not name:
            if not isinstance(handler, View):
                name = clean_string(handler.fn.__name__)
            else:
                name = clean_string(handler.__class__.__name__)

        # Handle middleware
        self.middleware = middleware or []
        self._middleware: List["Middleware"] = self.handle_middleware(
            handler=handler, base_middleware=self.middleware
        )
        super().__init__(
            path=self.path,
            handler=cast(Callable, handler),
            include_in_schema=include_in_schema,
            name=name,
            methods=self.methods,
            middleware=self._middleware,
            exception_handlers=exception_handlers,
        )
        """
        A "bridge" to a handler and router mapping functionality.
        Since the default Lilya Route handler does not understand the Esmerald handlers,
        the Gateway bridges both functionalities and adds an extra "flair" to be compliant with both class based views and decorated function views.
        """
        self._interceptors: Union[List["Interceptor"], "VoidType"] = Void
        self.name = name
        self.handler = cast("Callable", handler)
        self.dependencies = dependencies or {}
        self.interceptors: Sequence["Interceptor"] = interceptors or []
        self.permissions: Sequence["Permission"] = permissions or []  # type: ignore
        self.response_class = None
        self.response_cookies = None
        self.response_headers = None
        self.deprecated = deprecated
        self.parent = parent
        self.security = security
        self.tags = tags or []
        (handler.path_regex, handler.path_format, handler.param_convertors, _) = compile_path(
            self.path
        )
        self.operation_id = operation_id

        if self.is_handler(self.handler):  # type: ignore
            self.handler.name = self.name

            if self.operation_id or handler.operation_id is not None:
                handler_id = self.generate_operation_id(
                    name=self.name, handler=self.handler  # type: ignore
                )
                self.operation_id = f"{operation_id}_{handler_id}" if operation_id else handler_id

            elif not handler.operation_id:
                handler.operation_id = self.generate_operation_id(
                    name=self.name, handler=self.handler  # type: ignore
                )

    async def handle_dispatch(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """
        Handles the interception of messages and calls from the API.
        """
        await self.app(scope, receive, send)


class WebSocketGateway(LilyaWebSocketPath, Dispatcher, BaseMiddleware):
    """
    `WebSocketGateway` object class used by Esmerald routes.

    The WebSocketGateway act as a brigde between the router handlers and
    the main Esmerald routing system.

    Read more about [WebSocketGateway](https://esmerald.dev/routing/routes/#websocketgateway) and
    how to use it.

    **Example**

    ```python
    from esmerald import Esmerald. websocket

    @websocket()
    async def world_socket(socket: Websocket) -> None:
        await socket.accept()
        msg = await socket.receive_json()
        assert msg
        assert socket
        await socket.close()

    WebSocketGateway(path="/ws", handler=home)
    ```
    """

    __slots__ = (
        "_interceptors",
        "path",
        "handler",
        "name",
        "dependencies",
        "middleware",
        "exception_handlers",
        "interceptors",
        "permissions",
        "parent",
        "security",
        "tags",
    )

    def __init__(
        self,
        path: Annotated[
            Optional[str],
            Doc(
                """
                Relative path of the `WebSocketGateway`.
                The path can contain parameters in a dictionary like format
                and if the path is not provided, it will default to `/`.

                **Example**

                ```python
                WebSocketGateway()
                ```

                **Example with parameters**

                ```python
                WebSocketGateway(path="/{age: int}")
                ```
                """
            ),
        ] = None,
        *,
        handler: Annotated[
            Union["WebSocketHandler", View],
            Doc(
                """
            An instance of [handler](https://esmerald.dev/routing/handlers/#websocket-handler).
            """
            ),
        ],
        name: Annotated[
            Optional[str],
            Doc(
                """
                The name for the Gateway. The name can be reversed by `path_for()`.
                """
            ),
        ] = None,
        parent: Annotated[
            Optional["ParentType"],
            Doc(
                """
                Who owns the Gateway. If not specified, the application automatically it assign it.

                This is directly related with the [application levels](https://esmerald.dev/application/levels/).
                """
            ),
        ] = None,
        dependencies: Annotated[
            Optional["Dependencies"],
            Doc(
                """
                A dictionary of string and [Inject](https://esmerald.dev/dependencies/) instances enable application level dependency injection.
                """
            ),
        ] = None,
        middleware: Annotated[
            Optional[List["Middleware"]],
            Doc(
                """
                A list of middleware to run for every request. The middlewares of a Gateway will be checked from top-down or [Lilya Middleware](https://www.lilya.dev/middleware/) as they are both converted internally. Read more about [Python Protocols](https://peps.python.org/pep-0544/).
                """
            ),
        ] = None,
        interceptors: Annotated[
            Optional[Sequence["Interceptor"]],
            Doc(
                """
                A list of [interceptors](https://esmerald.dev/interceptors/) to serve the application incoming requests (HTTP and Websockets).
                """
            ),
        ] = None,
        permissions: Annotated[
            Optional[Sequence["Permission"]],
            Doc(
                """
                A list of [permissions](https://esmerald.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
            ),
        ] = None,
        exception_handlers: Annotated[
            Optional["ExceptionHandlerMap"],
            Doc(
                """
                A dictionary of [exception types](https://esmerald.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
            ),
        ] = None,
        is_from_router: Annotated[
            bool,
            Doc(
                """
                Used by the `.add_router()` function of the `Esmerald` class indicating if the
                Gateway is coming from a router.
                """
            ),
        ] = False,
    ) -> None:
        if not path:
            path = "/"
        if is_class_and_subclass(handler, View):
            handler = handler(parent=self)  # type: ignore

        if not is_from_router:
            self.path = clean_path(path + handler.path)
        else:
            self.path = clean_path(path)

        if not name:
            if not isinstance(handler, View):
                name = clean_string(handler.fn.__name__)
            else:
                name = clean_string(handler.__class__.__name__)

        # Handle middleware
        self.middleware = middleware or []
        self._middleware: List["Middleware"] = self.handle_middleware(
            handler=handler, base_middleware=self.middleware
        )
        self.is_middleware: bool = False

        super().__init__(
            path=self.path,
            handler=cast("Callable", handler),
            name=name,
            middleware=self._middleware,
            exception_handlers=exception_handlers,
        )
        """
        A "bridge" to a handler and router mapping functionality.
        Since the default Lilya Route handler does not understand the Esmerald handlers,
        the Gateway bridges both functionalities and adds an extra "flair" to be compliant with both class based views and decorated function views.
        """
        self._interceptors: Union[List["Interceptor"], "VoidType"] = Void
        self.handler = cast("Callable", handler)
        self.dependencies = dependencies or {}
        self.interceptors = interceptors or []
        self.permissions = permissions or []  # type: ignore
        self.include_in_schema = False
        self.parent = parent
        (handler.path_regex, handler.path_format, handler.param_convertors, _) = compile_path(
            self.path
        )

    async def handle_dispatch(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """
        Handles the interception of messages and calls from the API.
        if self._middleware:
            self.is_middleware = True
        """
        await self.app(scope, receive, send)


class WebhookGateway(LilyaPath, Dispatcher, GatewayUtil):
    """
    `WebhookGateway` object class used by Esmerald routes.

    The WebhookGateway act as a brigde between the webhook handlers and
    the main Esmerald routing system.

    Read more about [WebhookGateway](https://esmerald.dev/routing/webhooks/) and
    how to use it.

    !!! Note
        This is used for OpenAPI documentation purposes only.
    """

    __slots__ = (
        "_interceptors",
        "path",
        "handler",
        "name",
        "include_in_schema",
        "parent",
        "dependencies",
        "middleware",
        "exception_handlers",
        "interceptors",
        "permissions",
        "security",
        "tags",
    )

    def __init__(
        self,
        *,
        handler: Annotated[
            Union["WebhookHandler", View],
            Doc(
                """
                An instance of [handler](https://esmerald.dev/routing/webhooks/#handlers).
                """
            ),
        ],
        name: Annotated[
            Optional[str],
            Doc(
                """
                The name for the WebhookGateway.
                """
            ),
        ] = None,
        include_in_schema: Annotated[
            bool,
            Doc(
                """
                Boolean flag indicating if it should be added to the OpenAPI docs.
                """
            ),
        ] = True,
        parent: Annotated[
            Optional["ParentType"],
            Doc(
                """
                Who owns the Gateway. If not specified, the application automatically it assign it.

                This is directly related with the [application levels](https://esmerald.dev/application/levels/).
                """
            ),
        ] = None,
        deprecated: Annotated[
            Optional[bool],
            Doc(
                """
                Boolean flag for indicating the deprecation of the Gateway and to display it
                in the OpenAPI documentation..
                """
            ),
        ] = None,
        security: Annotated[
            Optional[Sequence["SecurityScheme"]],
            Doc(
                """
                Used by OpenAPI definition, the security must be compliant with the norms.
                Esmerald offers some out of the box solutions where this is implemented.

                The [Esmerald security](https://esmerald.dev/openapi/) is available to automatically used.

                The security can be applied also on a [level basis](https://esmerald.dev/application/levels/).

                For custom security objects, you **must** subclass
                `esmerald.openapi.security.base.HTTPBase` object.
                """
            ),
        ] = None,
        tags: Annotated[
            Optional[Sequence[str]],
            Doc(
                """
                A list of strings tags to be applied to the *path operation*.

                It will be added to the generated OpenAPI documentation.

                **Note** almost everything in Esmerald can be done in [levels](https://esmerald.dev/application/levels/), which means
                these tags on a Esmerald instance, means it will be added to every route even
                if those routes also contain tags.
                """
            ),
        ] = None,
    ) -> None:
        if is_class_and_subclass(handler, View):
            handler = handler(parent=self)  # type: ignore

        self.path = handler.path
        self.methods = getattr(handler, "http_methods", None)

        if not name:
            if not isinstance(handler, View):
                name = clean_string(handler.fn.__name__)
            else:
                name = clean_string(handler.__class__.__name__)

        self.handler = cast("Callable", handler)
        self.include_in_schema = include_in_schema

        self._interceptors: Union[List["Interceptor"], "VoidType"] = Void
        self.name = name
        self.dependencies: Any = {}
        self.interceptors: Sequence["Interceptor"] = []
        self.permissions: Sequence["Permission"] = []  # type: ignore
        self.middleware: Any = []
        self.exception_handlers: Any = {}
        self.response_class = None
        self.response_cookies = None
        self.response_headers = None
        self.deprecated = deprecated
        self.parent = parent
        self.security = security
        self.tags = tags or []
        (handler.path_regex, handler.path_format, handler.param_convertors, _) = compile_path(
            self.path
        )

        if self.is_handler(self.handler):  # type: ignore
            self.handler.name = self.name

            if not handler.operation_id:
                handler.operation_id = self.generate_operation_id(
                    name=self.name, handler=self.handler  # type: ignore
                )



# from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Union

from pydantic.dataclasses import dataclass
from pydantic.fields import AliasChoices, AliasPath, FieldInfo

from esmerald.enums import EncodingType, ParamType
from esmerald.typing import Undefined
from esmerald.utils.constants import IS_DEPENDENCY, SKIP_VALIDATION

_PyUndefined: Any = Undefined


class Param(FieldInfo):
    in_: ParamType

    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        annotation: Optional[Any] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        value_type: Any = _PyUndefined,
        header: Optional[str] = None,
        cookie: Optional[str] = None,
        query: Optional[str] = None,
        path: Optional[str] = None,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = _PyUndefined,
        allow_inf_nan: Optional[bool] = _PyUndefined,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        pattern: Optional[str] = None,
        examples: Optional[List[Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
    ) -> None:
        self.deprecated = deprecated
        self.examples = examples
        self.include_in_schema = include_in_schema
        self.allow_none = allow_none

        extra: Dict[str, Any] = {}
        extra.update(header=header)
        extra.update(cookie=cookie)
        extra.update(query=query)
        extra.update(path=path)
        extra.update(required=required)
        extra.update(content_encoding=content_encoding)
        extra.update(value_type=value_type)
        extra.update(examples=self.examples)
        extra.update(deprecated=self.deprecated)
        extra.update(include_in_schema=self.include_in_schema)
        extra.update(allow_none=self.allow_none)

        super().__init__(
            annotation=annotation,
            default=default,
            default_factory=default_factory,
            alias=alias,
            alias_priority=alias_priority,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_length=min_length,
            max_length=max_length,
            pattern=pattern,
            examples=examples,
            allow_inf_nan=allow_inf_nan,
            json_schema_extra=extra,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            max_digits=max_digits,
            strict=strict,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(annotation={self.annotation}, default={self.default})"


class Header(Param):
    in_ = ParamType.HEADER

    def __init__(
        self,
        *,
        default: Any = _PyUndefined,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        annotation: Optional[Any] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        value: Optional[str] = None,
        value_type: Any = _PyUndefined,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = _PyUndefined,
        allow_inf_nan: Optional[bool] = _PyUndefined,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        pattern: Optional[str] = None,
        examples: Optional[List[Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
    ) -> None:
        super().__init__(
            default=default,
            allow_none=allow_none,
            default_factory=default_factory,
            annotation=annotation,
            header=value,
            alias=alias,
            alias_priority=alias_priority,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            allow_inf_nan=allow_inf_nan,
            min_length=min_length,
            max_length=max_length,
            max_digits=max_digits,
            strict=strict,
            pattern=pattern,
            required=required,
            content_encoding=content_encoding,
            value_type=value_type,
            examples=examples,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
        )


class Cookie(Param):
    in_ = ParamType.COOKIE

    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        annotation: Optional[Any] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        value_type: Any = _PyUndefined,
        value: Optional[str] = None,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = _PyUndefined,
        allow_inf_nan: Optional[bool] = _PyUndefined,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        pattern: Optional[str] = None,
        examples: Optional[List[Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
    ) -> None:
        super().__init__(
            default=default,
            allow_none=allow_none,
            default_factory=default_factory,
            annotation=annotation,
            cookie=value,
            alias=alias,
            alias_priority=alias_priority,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            allow_inf_nan=allow_inf_nan,
            min_length=min_length,
            max_length=max_length,
            max_digits=max_digits,
            strict=strict,
            pattern=pattern,
            required=required,
            content_encoding=content_encoding,
            value_type=value_type,
            examples=examples,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
        )


class Query(Param):
    in_ = ParamType.QUERY

    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        annotation: Optional[Any] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        value_type: Any = _PyUndefined,
        value: Optional[str] = None,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = _PyUndefined,
        allow_inf_nan: Optional[bool] = _PyUndefined,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        pattern: Optional[str] = None,
        examples: Optional[List[Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
    ) -> None:
        super().__init__(
            default=default,
            allow_none=allow_none,
            default_factory=default_factory,
            annotation=annotation,
            query=value,
            alias=alias,
            alias_priority=alias_priority,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            allow_inf_nan=allow_inf_nan,
            min_length=min_length,
            max_length=max_length,
            max_digits=max_digits,
            strict=strict,
            pattern=pattern,
            required=required,
            content_encoding=content_encoding,
            value_type=value_type,
            examples=examples,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
        )


class Path(Param):
    in_ = ParamType.PATH

    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        annotation: Optional[Any] = None,
        value_type: Any = _PyUndefined,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        pattern: Optional[str] = None,
        examples: Optional[List[Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
    ) -> None:
        super().__init__(
            default=default,
            allow_none=allow_none,
            default_factory=default_factory,
            annotation=annotation,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            min_length=min_length,
            max_length=max_length,
            max_digits=max_digits,
            strict=strict,
            pattern=pattern,
            required=required,
            content_encoding=content_encoding,
            value_type=value_type,
            examples=examples,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
        )


class Body(FieldInfo):
    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        annotation: Optional[Any] = None,
        media_type: Union[str, EncodingType] = EncodingType.JSON,
        content_encoding: Optional[str] = None,
        title: Optional[str] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        description: Optional[str] = None,
        embed: Optional[bool] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = _PyUndefined,
        allow_inf_nan: Optional[bool] = _PyUndefined,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        pattern: Optional[str] = None,
        examples: Optional[List[Any]] = None,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
        include_in_schema: bool = True,
    ) -> None:
        extra: Dict[str, Any] = {}
        self.media_type = media_type
        self.content_encoding = content_encoding
        self.examples = examples
        self.allow_none = allow_none
        self.include_in_schema = include_in_schema

        extra.update(media_type=self.media_type)
        extra.update(content_encoding=self.content_encoding)
        extra.update(embed=embed)
        extra.update(allow_none=allow_none)

        super().__init__(
            default=default,
            default_factory=default_factory,
            annotation=annotation,
            title=title,
            alias=alias,
            alias_priority=alias_priority,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            max_digits=max_digits,
            allow_inf_nan=allow_inf_nan,
            min_length=min_length,
            max_length=max_length,
            pattern=pattern,
            json_schema_extra=extra,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            frozen=frozen,
            init_var=init_var,
            strict=strict,
            kw_only=kw_only,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(annotation={self.annotation}, default={self.default})"


class Form(Body):
    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        annotation: Optional[Any] = None,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        allow_none: Optional[bool] = True,
        media_type: Union[str, EncodingType] = EncodingType.URL_ENCODED,
        content_encoding: Optional[str] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        pattern: Optional[str] = None,
        examples: Optional[List[Any]] = None,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
        include_in_schema: bool = True,
    ) -> None:
        super().__init__(
            default=default,
            annotation=annotation,
            allow_none=allow_none,
            default_factory=default_factory,
            embed=True,
            media_type=media_type,
            content_encoding=content_encoding,
            alias=alias,
            alias_priority=alias_priority,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            min_length=min_length,
            max_length=max_length,
            pattern=pattern,
            examples=examples,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
            include_in_schema=include_in_schema,
            max_digits=max_digits,
            strict=strict,
        )


class File(Form):
    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        annotation: Optional[Any] = None,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        media_type: Union[str, EncodingType] = EncodingType.MULTI_PART,
        content_encoding: Optional[str] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        pattern: Optional[str] = None,
        examples: Optional[List[Any]] = None,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
        include_in_schema: bool = True,
    ) -> None:
        super().__init__(
            default=default,
            annotation=annotation,
            allow_none=allow_none,
            default_factory=default_factory,
            media_type=media_type,
            content_encoding=content_encoding,
            alias=alias,
            alias_priority=alias_priority,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            min_length=min_length,
            max_length=max_length,
            pattern=pattern,
            examples=examples,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            max_digits=max_digits,
            strict=strict,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
            include_in_schema=include_in_schema,
        )


class Injects(FieldInfo):
    """
    Creates a FieldInfo class with extra parameters.
    This is used for dependencies and to inject them.

    **Example**

    ```python
    @get(dependencies={"value": Inject(lambda: 13)})
    def myview(value: Injects()):
        return {"value": value}
    ```
    """

    def __init__(
        self,
        default: Any = Undefined,
        skip_validation: bool = False,
        allow_none: bool = True,
    ) -> None:
        self.allow_none = allow_none
        self.extra: Dict[str, Any] = {
            IS_DEPENDENCY: True,
            SKIP_VALIDATION: skip_validation,
            "allow_none": self.allow_none,
        }
        super().__init__(default=default, json_schema_extra=self.extra)


@dataclass
class DirectInject:  # pragma: no cover
    def __init__(
        self,
        dependency: Optional[Callable[..., Any]] = None,
        *,
        use_cache: bool = True,
        allow_none: bool = True,
    ) -> None:
        self.dependency = dependency
        self.use_cache = use_cache
        self.allow_none = allow_none

    def __hash__(self) -> int:
        values: Dict[str, Any] = {}
        for key, value in self.__dict__.items():
            values[key] = None
            if isinstance(value, (list, set)):
                values[key] = tuple(value)
            else:
                values[key] = value
        return hash((type(self),) + tuple(values))



from http import HTTPStatus
from typing import Any, Dict, List, Optional, Tuple, Union, cast

from lilya import status
from lilya.exceptions import (
    EnvironmentError as EnvironmentError,
    HTTPException as LilyaHTTPException,
    ImproperlyConfigured as ImproperlyConfigured,
    LilyaException as LilyaException,
    MethodNotAllowed as MethodNotAllowed,
    NotAuthorized as NotAuthorized,
    NotFound as NotFound,
    PermissionDenied as PermissionDenied,
    TemplateNotFound as TemplateNotFound,
    WebSocketException as WebSocketException,
)
from typing_extensions import Annotated, Doc

from esmerald.contrib.encoding import force_str


def _get_error_details(data: Any) -> Any:
    if isinstance(data, (list, tuple)):
        return [_get_error_details(item) for item in data]

    elif isinstance(data, dict):
        return {key: _get_error_details(value) for key, value in data.items()}

    text = force_str(data)
    return ErrorDetail(text)


class ErrorDetail(str):
    def __new__(cls, string: str) -> "ErrorDetail":
        self = super().__new__(cls, string)
        return self

    def __repr__(self) -> str:
        return "ErrorDetail(string=%r)" % (str(self))

    def __hash__(self) -> int:
        return hash(str(self))


class EsmeraldAPIException(LilyaException): ...


class HTTPException(LilyaHTTPException, EsmeraldAPIException):
    """
    Base of all `Esmerald` execeptions.

    !!! Tip
        For an implementation of a custom exception that you need
        to be thrown by Esmerald, it is advised to subclass `HTTPException`.
    """

    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

    def __init__(
        self,
        *args: Annotated[
            Any,
            Doc(
                """
                The args passed to the exception.
                """
            ),
        ],
        detail: Annotated[
            Optional[str],
            Doc(
                """
                A string text with the details of the error being thrown.
                """
            ),
        ] = None,
        status_code: Annotated[
            Optional[int],
            Doc(
                """
                An integer with the status code to be raised.
                """
            ),
        ] = None,
        headers: Annotated[
            Optional[Dict[str, Any]],
            Doc(
                """
                Any python dictionary containing headers.
                """
            ),
        ] = None,
        **extra: Annotated[
            Any,
            Doc(
                """
                Any extra information used by the exception.
                """
            ),
        ],
    ) -> None:
        detail = detail or getattr(self, "detail", None)
        status_code = status_code or getattr(self, "status_code", None)
        if not detail:
            detail = args[0] if args else HTTPStatus(status_code or self.status_code).phrase
            args = args[1:]
        super().__init__(status_code=status_code, detail=detail, headers=headers)
        self.detail = detail
        self.headers = headers
        self.args = (f"{self.status_code}: {self.detail}", *args)
        self.extra = extra

    def __repr__(self) -> str:  # pragma: no cover
        return f"<{self.status_code}: {self.__class__.__name__} />"


class ImproperlyMiddlewareConfigured(ImproperlyConfigured): ...


class NotAuthenticated(HTTPException):
    status_code = status.HTTP_401_UNAUTHORIZED
    detail = "Authentication credentials were not provided."


class ValidationErrorException(HTTPException, ValueError):
    status_code = status.HTTP_400_BAD_REQUEST


class InternalServerError(HTTPException):
    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR


class ServiceUnavailable(HTTPException):
    status_code = status.HTTP_503_SERVICE_UNAVAILABLE


class MissingDependency(EsmeraldAPIException, ImportError): ...


class OpenAPIException(ImproperlyConfigured): ...


class AuthenticationError(HTTPException):
    status_code = status.HTTP_401_UNAUTHORIZED


class ValidationError(HTTPException):
    """
    Provides a more detailed error message for validation errors
    when thrown by the application.
    """

    status_code = status.HTTP_400_BAD_REQUEST
    detail = "Validation error."

    def __init__(
        self,
        detail: Union[str, List[str], Dict[str, Any], Tuple[str]] = None,
        status_code: Annotated[
            Optional[int],
            Doc(
                """
                An integer with the status code to be raised.
                """
            ),
        ] = None,
        headers: Annotated[
            Optional[Dict[str, Any]],
            Doc(
                """
                Any python dictionary containing headers.
                """
            ),
        ] = None,
        **extra: Annotated[
            Any,
            Doc(
                """
                Any extra information used by the exception.
                """
            ),
        ],
    ) -> None:
        if isinstance(detail, tuple):
            detail = list(detail)
        elif not isinstance(detail, dict) and not isinstance(detail, list):
            detail = [detail]

        detail = _get_error_details(detail)
        super().__init__(
            status_code=status_code, detail=cast(str, detail), headers=headers, **extra
        )


ExceptionErrorMap = Union[
    HTTPException,
    NotAuthenticated,
    PermissionDenied,
    NotAuthorized,
    NotFound,
    MethodNotAllowed,
    InternalServerError,
    ServiceUnavailable,
]



from lilya.cli.base import BaseDirective as LilyaBaseDirective

import esmerald
from esmerald.core.terminal.print import Print
from esmerald.parsers import ArbitraryExtraBaseModel

printer = Print()


class BaseDirective(ArbitraryExtraBaseModel, LilyaBaseDirective):
    """The base class from which all directives derive."""

    def get_version(self) -> str:
        """
        Returns the current version of Esmerald.
        """
        return esmerald.__version__



from inspect import Signature as InspectSignature

from esmerald.typing import Undefined

UNDEFINED = {Undefined, InspectSignature.empty}
CLASS_SPECIAL_WORDS = {"self", "cls"}
VALIDATION_NAMES = {"request", "socket", "state"}



from http import HTTPStatus
from inspect import Signature
from typing import TYPE_CHECKING, Any, Dict, Union, cast

from typing_extensions import get_args, get_origin

from esmerald.datastructures import File, Redirect, Stream, Template
from esmerald.enums import MediaType
from esmerald.openapi._internal import InternalResponse
from esmerald.responses import Response as EsmeraldResponse

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.routing.router import HTTPHandler
    from esmerald.typing import AnyCallable


def create_internal_response(
    handler: Union["HTTPHandler", Any]
) -> InternalResponse:  # pragma: no cover
    signature = Signature.from_callable(cast("AnyCallable", handler.fn))
    default_descriptions: Dict[Any, str] = {
        Stream: "Stream Response",
        Redirect: "Redirect Response",
        File: "File Download",
    }

    description = (
        handler.response_description
        or default_descriptions.get(signature.return_annotation)
        or HTTPStatus(handler.status_code).description
    )

    internal_response = InternalResponse(
        description=description, signature=signature, return_annotation=signature.return_annotation
    )

    if signature.return_annotation not in {
        signature.empty,
        None,
        Redirect,
        File,
        Stream,
    }:
        if signature.return_annotation is Template:
            internal_response.return_annotation = str
            internal_response.media_type = MediaType.HTML
        elif get_origin(signature.return_annotation) is EsmeraldResponse:
            internal_response.return_annotation = get_args(signature.return_annotation)[0] or Any
            internal_response.media_type = handler.content_media_type
        else:
            internal_response.media_type = MediaType.JSON

        internal_response.encoding = handler.content_encoding

    elif signature.return_annotation is Redirect:
        internal_response.media_type = MediaType.JSON
    elif signature.return_annotation in (File, Stream):
        internal_response.media_type = handler.content_media_type
        internal_response.encoding = handler.content_encoding or MediaType.OCTET
    else:
        internal_response.media_type = handler.content_media_type
        internal_response.encoding = handler.content_encoding
    return internal_response



from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Literal

from esmerald.openapi.enums import APIKeyIn, SecuritySchemeType
from esmerald.openapi.schemas.v3_1_0.contact import Contact as Contact
from esmerald.openapi.schemas.v3_1_0.discriminator import Discriminator as Discriminator
from esmerald.openapi.schemas.v3_1_0.encoding import Encoding as Encoding
from esmerald.openapi.schemas.v3_1_0.example import Example as Example
from esmerald.openapi.schemas.v3_1_0.external_documentation import (
    ExternalDocumentation as ExternalDocumentation,
)
from esmerald.openapi.schemas.v3_1_0.header import Header as Header
from esmerald.openapi.schemas.v3_1_0.info import Info as Info
from esmerald.openapi.schemas.v3_1_0.license import License as License
from esmerald.openapi.schemas.v3_1_0.link import Link as Link
from esmerald.openapi.schemas.v3_1_0.media_type import MediaType as MediaType
from esmerald.openapi.schemas.v3_1_0.oauth_flow import OAuthFlow as OpenOAuthFlow
from esmerald.openapi.schemas.v3_1_0.oauth_flows import OAuthFlows as OAuthFlows
from esmerald.openapi.schemas.v3_1_0.operation import Operation as Operation
from esmerald.openapi.schemas.v3_1_0.parameter import Parameter as Parameter
from esmerald.openapi.schemas.v3_1_0.path_item import PathItem as PathItem
from esmerald.openapi.schemas.v3_1_0.paths import Paths as Paths
from esmerald.openapi.schemas.v3_1_0.reference import Reference as Reference
from esmerald.openapi.schemas.v3_1_0.request_body import RequestBody as RequestBody
from esmerald.openapi.schemas.v3_1_0.response import Response as Response
from esmerald.openapi.schemas.v3_1_0.schema import Schema as Schema
from esmerald.openapi.schemas.v3_1_0.security_requirement import (
    SecurityRequirement as SecurityRequirement,
)
from esmerald.openapi.schemas.v3_1_0.security_scheme import SecurityScheme as SecurityScheme
from esmerald.openapi.schemas.v3_1_0.server import Server as Server
from esmerald.openapi.schemas.v3_1_0.server_variable import ServerVariable as ServerVariable
from esmerald.openapi.schemas.v3_1_0.tag import Tag as Tag
from esmerald.openapi.schemas.v3_1_0.xml import XML as XML


class APIKey(SecurityScheme):
    type: Literal["apiKey", "http", "mutualTLS", "oauth2", "openIdConnect"] = Field(
        default=SecuritySchemeType.apiKey,
        alias="type",
    )
    param_in: APIKeyIn = Field(alias="in")
    name: str


class HTTPBase(SecurityScheme):
    type: Literal["apiKey", "http", "mutualTLS", "oauth2", "openIdConnect"] = Field(
        default=SecuritySchemeType.http,
        alias="type",
    )
    scheme: str


class HTTPBearer(HTTPBase):
    scheme: Literal["bearer"] = "bearer"
    bearerFormat: Optional[str] = None


class OAuthFlow(OpenOAuthFlow):
    scopes: Dict[str, str] = {}


class OAuth2(SecurityScheme):
    type: Literal["apiKey", "http", "mutualTLS", "oauth2", "openIdConnect"] = Field(
        default=SecuritySchemeType.oauth2, alias="type"
    )
    flows: OAuthFlows


class OpenIdConnect(SecurityScheme):
    type: Literal["apiKey", "http", "mutualTLS", "oauth2", "openIdConnect"] = Field(
        default=SecuritySchemeType.openIdConnect, alias="type"
    )
    openIdConnectUrl: str


class SecurityBase(BaseModel):
    model_config = ConfigDict(extra="allow")

    type: SecuritySchemeType = Field(alias="type")
    description: Optional[str] = None


SecuritySchemeUnion = Union[APIKey, HTTPBase, OAuth2, OpenIdConnect, HTTPBearer]


class Components(BaseModel):
    schemas: Optional[Dict[str, Union[Schema, Reference]]] = None
    responses: Optional[Dict[str, Union[Response, Reference]]] = None
    parameters: Optional[Dict[str, Union[Parameter, Reference]]] = None
    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    requestBodies: Optional[Dict[str, Union[RequestBody, Reference]]] = None
    headers: Optional[Dict[str, Union[Header, Reference]]] = None
    securitySchemes: Optional[Dict[str, Union[SecurityScheme, Reference]]] = None
    links: Optional[Dict[str, Union[Link, Reference]]] = None
    callbacks: Optional[Dict[str, Union[Dict[str, PathItem], Reference, Any]]] = None
    pathItems: Optional[Dict[str, Union[PathItem, Reference]]] = None

    model_config = ConfigDict(extra="allow")


class OpenAPI(BaseModel):
    openapi: str
    info: Info
    jsonSchemaDialect: Optional[str] = None
    servers: Optional[List[Dict[str, Union[str, Any]]]] = None
    paths: Optional[Dict[str, Union[PathItem, Any]]] = None
    webhooks: Optional[Dict[str, Union[PathItem, Reference]]] = None
    components: Optional[Components] = None
    security: Optional[List[Dict[str, List[str]]]] = None
    tags: Optional[List[str]] = None
    externalDocs: Optional[ExternalDocumentation] = None
    model_config = ConfigDict(extra="allow")



import click

from esmerald.core.directives.exceptions import DirectiveError
from esmerald.core.directives.templates import TemplateDirective
from esmerald.core.terminal import Print

printer = Print()


@click.option("-v", "--verbosity", default=1, type=int, help="Displays the files generated")
@click.option(
    "--deployment-folder-name",
    default="deployment",
    show_default=True,
    type=str,
    help="The name of the folder for the deployment files.",
)
@click.argument("name", type=str)
@click.command(name="createdeployment")
def create_deployment(name: str, verbosity: int, deployment_folder_name: str) -> None:
    """
    Generates the scaffold for the deployment of an Esmerald application.

    The scaffold contains the configurations for docker, nginx, supervisor and gunicorn.

    The configurations should be adapted accordingly.
    The parameter <NAME> corresponds to the name of the
    project where the deployment should be placed.

    How to run: `esmerald createdeployment <NAME>`

    Example: `esmerald createdeployment myproject`
    """
    options = {
        "verbosity": verbosity,
        "deployment_folder_name": deployment_folder_name,
    }
    directive = TemplateDirective()

    try:
        directive.handle("deployment", name=name, **options)
        printer.write_success(f"Deployment for {name} generated successfully!")
    except DirectiveError as e:
        printer.write_error(str(e))



from typing import Union

from pydantic import BaseModel, ConfigDict, constr, field_validator
from typing_extensions import Annotated, Doc, Literal

from esmerald.datastructures import Secret

SECONDS_IN_A_DAY: Annotated[
    int,
    Doc(
        """
        Total seconds in a day.
        """
    ),
] = (
    60 * 60 * 24
)


class SessionConfig(BaseModel):
    """
    An instance of [SessionConfig](https://esmerald.dev/configurations/session/).

    This configuration is passed to the [SessionMiddleware](https://esmerald.dev/middleware/middleware/#sessionmiddleware) and enables the middleware.

    **Example**

    ```python
    from esmerald import Esmerald
    from esmerald.config import SessionConfig

    session_config = SessionConfig(
        secret_key=settings.secret_key,
        session_cookie="session",
    )

    app = Esmerald(session_config=session_config)
    ```
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    secret_key: Annotated[
        Union[str, Secret],
        Doc(
            """
            The string used for the encryption/decryption and used to create an HMAC to sign.

            !!! Tip
                It is advised to use the same secret as the one in the settings to make it consistent.
            """
        ),
    ]
    path: Annotated[
        str,
        Doc(
            """
            The path of the cookie.
            """
        ),
    ] = "/"
    session_cookie: Annotated[  # type: ignore
        constr(min_length=1, max_length=256),
        Doc(
            """
            The name for the session cookie.
            """
        ),
    ] = "session"
    max_age: Annotated[
        int,
        Doc(
            """
            The number in seconds until the cookie expires.
            """
        ),
    ] = (
        SECONDS_IN_A_DAY * 180
    )
    https_only: Annotated[
        bool,
        Doc(
            """
            Boolean if set enforces the session cookie to be httpsOnly.
            """
        ),
    ] = False
    same_site: Annotated[
        Literal["lax", "strict", "none"],
        Doc(
            """
            Level of restriction for the session cookie.

            Learn more about the [same site](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie/SameSite).
            """
        ),
    ] = "lax"

    @field_validator("secret_key")
    def validate_secret(
        cls,
        value: Annotated[
            Secret,
            Doc(
                """
                The string secret that will be evaluated.
                """
            ),
        ],
    ) -> Secret:
        if len(value) not in [16, 24, 32]:
            raise ValueError("secret length must be 16 (128 bit), 24 (192 bit) or 32 (256 bit)")
        return value



from typing import Union

from lilya import status
from lilya.exceptions import HTTPException as LilyaHTTPException
from lilya.requests import Request
from lilya.responses import Response as LilyaResponse
from orjson import loads
from pydantic import ValidationError

from esmerald.enums import MediaType
from esmerald.exceptions import (
    ExceptionErrorMap,
    HTTPException,
    ImproperlyConfigured,
)
from esmerald.responses import JSONResponse, Response


async def http_exception_handler(
    request: Request, exc: Union[HTTPException, LilyaHTTPException]
) -> Union[JSONResponse, Response]:  # pragma: no cover
    """
    Default exception handler for LilyaHTTPException and Esmerald HTTPException.
    """

    extra = getattr(exc, "extra", None)
    headers = getattr(exc, "headers", None)

    if exc.status_code in {204, 304}:
        return JSONResponse(None, status_code=exc.status_code, headers=headers)

    if headers and not extra:
        return JSONResponse({"detail": exc.detail}, status_code=exc.status_code, headers=headers)
    elif headers and extra:
        return JSONResponse(
            {"detail": exc.detail, "extra": extra},
            status_code=exc.status_code,
            headers=headers,
        )
    elif not headers and extra:
        return JSONResponse({"detail": exc.detail, "extra": extra}, status_code=exc.status_code)
    else:
        return JSONResponse({"detail": exc.detail}, status_code=exc.status_code)


async def validation_error_exception_handler(
    request: Request, exc: ValidationError
) -> JSONResponse:  # pragma: no cover
    extra = getattr(exc, "extra", None)
    status_code = status.HTTP_400_BAD_REQUEST

    if extra:
        errors_extra = exc.extra.get("extra", {})
        return JSONResponse(
            {"detail": exc.detail, "errors": errors_extra},
            status_code=status_code,
        )
    else:
        return JSONResponse(
            {"detail": exc.detail},
            status_code=status_code,
        )


async def http_error_handler(
    _: Request, exc: ExceptionErrorMap
) -> JSONResponse:  # pragma: no cover
    return JSONResponse({"detail": exc.detail}, status_code=exc.status_code)


async def improperly_configured_exception_handler(
    request: Request, exc: ImproperlyConfigured
) -> LilyaResponse:  # pragma: no cover
    """
    When an ImproperlyConfiguredException is raised.
    """
    status_code = (
        exc.status_code
        if isinstance(exc, LilyaHTTPException)
        else status.HTTP_500_INTERNAL_SERVER_ERROR
    )
    if not status_code:
        status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

    content = {"detail": exc.detail}
    if exc.extra:
        content.update({"extra": exc.extra})  # type: ignore[dict-item]
    headers = exc.headers if isinstance(exc, (HTTPException, LilyaHTTPException)) else None

    return Response(
        media_type=MediaType.JSON,
        content=content,
        status_code=status_code,
        headers=headers,
    )


async def pydantic_validation_error_handler(
    request: Request, exc: ValidationError
) -> JSONResponse:  # pragma: no cover
    """
    This handler is to be used when a pydantic validation error is triggered during the logic
    of a code block and not the definition of a handler.

    This is different from validation_error_exception_handler
    """
    status_code = status.HTTP_422_UNPROCESSABLE_ENTITY
    return JSONResponse({"detail": loads(exc.json())}, status_code=status_code)


async def value_error_handler(
    request: Request, exc: ValueError
) -> JSONResponse:  # pragma: no cover
    """
    Simple handler that manages all the ValueError exceptions thrown to the user properly
    formatted.
    """
    status_code = status.HTTP_400_BAD_REQUEST
    details = loads(exc.json()) if hasattr(exc, "json") else exc.args[0]
    return JSONResponse({"detail": details}, status_code=status_code)



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Union, cast

from lilya import status
from typing_extensions import Annotated, Doc

from esmerald.enums import HttpMethod, MediaType
from esmerald.exceptions import ImproperlyConfigured
from esmerald.openapi.datastructures import OpenAPIResponse
from esmerald.permissions.types import Permission
from esmerald.routing.router import HTTPHandler, WebSocketHandler
from esmerald.types import (
    BackgroundTaskType,
    Dependencies,
    ExceptionHandlerMap,
    Middleware,
    ResponseCookies,
    ResponseHeaders,
    ResponseType,
)
from esmerald.utils.constants import AVAILABLE_METHODS

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.openapi.schemas.v3_1_0 import SecurityScheme


SUCCESSFUL_RESPONSE = "Successful response"


def get(
    path: Annotated[
        Optional[str],
        Doc(
            """
                Relative path of the `handler`.
                The path can contain parameters in a dictionary like format
                and if the path is not provided, it will default to `/`.

                **Example**

                ```python
                @get()
                ```

                **Example with parameters**

                ```python
                @get(path="/{age: int}")
                ```
                """
        ),
    ] = None,
    *,
    summary: Annotated[
        Optional[str],
        Doc(
            """
                The summary of the handler. This short summary is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.

                **Example**

                ```python
                from esmerald import get


                @get(summary="Black Window joining Pretenders")
                async def get_joiners() -> None:
                    ...
                ```
                """
        ),
    ] = None,
    description: Annotated[
        Optional[str],
        Doc(
            """
                The description of the Esmerald application/API. This description is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.

                **Example**

                ```python
                from esmerald import get


                @get(description=...)
                async def get_joiners() -> None:
                    ...
                """
        ),
    ] = None,
    status_code: Annotated[
        Optional[int],
        Doc(
            """
                An integer indicating the status code of the handler.

                This can be achieved by passing directly the value or
                by using the `esmerald.status` or even the `lilya.status`.
                """
        ),
    ] = status.HTTP_200_OK,
    content_encoding: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content encoding of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    content_media_type: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content media type of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    include_in_schema: Annotated[
        bool,
        Doc(
            """
                Boolean flag indicating if it should be added to the OpenAPI docs.
                """
        ),
    ] = True,
    background: Annotated[
        Optional["BackgroundTaskType"],
        Doc(
            """
                An instance of an `esmerald.background.BackgroundTask` or `esmerald.background.BackgroundTasks` to be passed onto the handler.

                Read more about [Background tasks](https://esmerald.dev/background-tasks/) to
                understand what can be done.
                """
        ),
    ] = None,
    dependencies: Annotated[
        Optional["Dependencies"],
        Doc(
            """
                A dictionary of string and [Inject](https://esmerald.dev/dependencies/) instances enable application level dependency injection.
                """
        ),
    ] = None,
    exception_handlers: Annotated[
        Optional["ExceptionHandlerMap"],
        Doc(
            """
                A dictionary of [exception types](https://esmerald.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
        ),
    ] = None,
    middleware: Annotated[
        Optional[List["Middleware"]],
        Doc(
            """
                A list of middleware to run for every request. The middlewares of an Include will be checked from top-down or [Lilya Middleware](https://www.lilya.dev/middleware/) as they are both converted internally. Read more about [Python Protocols](https://peps.python.org/pep-0544/).
                """
        ),
    ] = None,
    permissions: Annotated[
        Optional[List["Permission"]],
        Doc(
            """
                A list of [permissions](https://esmerald.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
        ),
    ] = None,
    media_type: Annotated[
        Union[MediaType, str],
        Doc(
            """
                The default `media-type` used by the handler.
                """
        ),
    ] = MediaType.JSON,
    response_class: Annotated[
        Optional["ResponseType"],
        Doc(
            """
                Response class to be used within the
                handler application.

                Read more about the [Responses](https://esmerald.dev/responses/) and how
                to use them.

                **Example**

                ```python
                from esmerald import get, JSONResponse

                @get(response_class=JSONResponse)
                ```
                """
        ),
    ] = None,
    response_cookies: Annotated[
        Optional["ResponseCookies"],
        Doc(
            """
                A sequence of `esmerald.datastructures.Cookie` objects.

                Read more about the [Cookies](https://esmerald.dev/extras/cookie-fields/?h=responsecook#cookie-from-response-cookies).

                **Example**

                ```python
                from esmerald import get
                from esmerald.datastructures import Cookie

                response_cookies=[
                    Cookie(
                        key="csrf",
                        value="CIwNZNlR4XbisJF39I8yWnWX9wX4WFoz",
                        max_age=3000,
                        httponly=True,
                    )
                ]

                @get(response_cookies=response_cookies)
                ```
                """
        ),
    ] = None,
    response_headers: Annotated[
        Optional["ResponseHeaders"],
        Doc(
            """
                A mapping of `esmerald.datastructures.ResponseHeader` objects.

                Read more about the [ResponseHeader](https://esmerald.dev/extras/header-fields/#response-headers).

                **Example**

                ```python
                from esmerald import get
                from esmerald.datastructures import ResponseHeader

                response_headers={
                    "authorize": ResponseHeader(value="granted")
                }

                @get(response_headers=response_headers)
                ```
                """
        ),
    ] = None,
    tags: Annotated[
        Optional[Sequence[str]],
        Doc(
            """
                A list of strings tags to be applied to the *path operation*.

                It will be added to the generated OpenAPI documentation.

                **Note** almost everything in Esmerald can be done in [levels](https://esmerald.dev/application/levels/), which means
                these tags on a Esmerald instance, means it will be added to every route even
                if those routes also contain tags.

                **Example**

                ```python
                from esmerald import get

                @get(tags=["application"])
                ```
                """
        ),
    ] = None,
    deprecated: Annotated[
        Optional[bool],
        Doc(
            """
                Boolean flag indicating if the handler
                should be deprecated in the OpenAPI documentation.

                **Example**

                ```python
                from esmerald import get

                @get(deprecated=True)
                ```
                """
        ),
    ] = None,
    security: Annotated[
        Optional[List["SecurityScheme"]],
        Doc(
            """
                Used by OpenAPI definition, the security must be compliant with the norms.
                Esmerald offers some out of the box solutions where this is implemented.

                The [Esmerald security](https://esmerald.dev/openapi/) is available to automatically used.

                The security can be applied also on a [level basis](https://esmerald.dev/application/levels/).

                For custom security objects, you **must** subclass
                `esmerald.openapi.security.base.HTTPBase` object.

                **Example**

                ```python
                from esmerald import get
                from esmerald.openapi.security.http import Bearer

                @get(security=[Bearer()])
                ```
                """
        ),
    ] = None,
    operation_id: Annotated[
        Optional[str],
        Doc(
            """
                The unique identifier of the `handler`. This acts as a unique ID
                for the OpenAPI documentation.

                !!! Tip
                    Usually you don't need this as Esmerald handles it automatically
                    but it is here if you want to add your own.
                """
        ),
    ] = None,
    response_description: Annotated[
        Optional[str],
        Doc(
            """
                A description of the response. This is used for OpenAPI documentation
                purposes only and accepts all the docstrings including `markdown` format.
                """
        ),
    ] = SUCCESSFUL_RESPONSE,
    responses: Annotated[
        Optional[Dict[int, OpenAPIResponse]],
        Doc(
            """
                Additional responses that are handled by the handler and need to be described
                in the OpenAPI documentation.

                The `responses` is a dictionary like object where the first parameter is an
                `integer` and the second is an instance of an [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object.


                Read more about [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object and how to use it.


                **Example**

                ```python
                from esmerald import get
                from esmerald.openapi.datastructures import OpenAPIResponse
                from pydantic import BaseModel

                class Power(BaseModel):
                    name: str
                    description: str


                class Error(BaseModel):
                    detail: str


                @get(path='/read', responses={
                        200: OpenAPIResponse(model=Power, description=...)
                        400: OpenAPIResponse(model=Error, description=...)
                    }
                )
                async def create() -> Union[None, ItemOut]:
                    ...
                ```
                """
        ),
    ] = None,
) -> HTTPHandler:
    """
    Handler responsible for the HTTP method `get` and
    all of its operatations.

    **Example**

    ```python
    from esmerald import get


    @get()
    async def get() -> str:
        return "Hello, World!"
    ```
    """

    def wrapper(func: Any) -> HTTPHandler:
        handler = HTTPHandler(
            path=path,
            methods=[HttpMethod.GET],
            summary=summary,
            description=description,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            response_description=response_description,
            responses=responses,
        )
        handler.fn = func
        handler.handler = func
        handler.__type__ = HttpMethod.GET.value
        handler.validate_handler()
        return handler

    return cast(HTTPHandler, wrapper)


def head(
    path: Annotated[
        Optional[str],
        Doc(
            """
                Relative path of the `handler`.
                The path can contain parameters in a dictionary like format
                and if the path is not provided, it will default to `/`.
                """
        ),
    ] = None,
    *,
    summary: Annotated[
        Optional[str],
        Doc(
            """
                The summary of the handler. This short summary is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.
                """
        ),
    ] = None,
    description: Annotated[
        Optional[str],
        Doc(
            """
                The description of the Esmerald application/API. This description is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.
                """
        ),
    ] = None,
    status_code: Annotated[
        Optional[int],
        Doc(
            """
                An integer indicating the status code of the handler.

                This can be achieved by passing directly the value or
                by using the `esmerald.status` or even the `lilya.status`.
                """
        ),
    ] = status.HTTP_200_OK,
    content_encoding: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content encoding of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    content_media_type: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content media type of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    include_in_schema: Annotated[
        bool,
        Doc(
            """
                Boolean flag indicating if it should be added to the OpenAPI docs.
                """
        ),
    ] = True,
    background: Annotated[
        Optional["BackgroundTaskType"],
        Doc(
            """
                An instance of an `esmerald.background.BackgroundTask` or `esmerald.background.BackgroundTasks` to be passed onto the handler.

                Read more about [Background tasks](https://esmerald.dev/background-tasks/) to
                understand what can be done.
                """
        ),
    ] = None,
    dependencies: Annotated[
        Optional["Dependencies"],
        Doc(
            """
                A dictionary of string and [Inject](https://esmerald.dev/dependencies/) instances enable application level dependency injection.
                """
        ),
    ] = None,
    exception_handlers: Annotated[
        Optional["ExceptionHandlerMap"],
        Doc(
            """
                A dictionary of [exception types](https://esmerald.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
        ),
    ] = None,
    middleware: Annotated[
        Optional[List["Middleware"]],
        Doc(
            """
                A list of middleware to run for every request. The middlewares of an Include will be checked from top-down or [Lilya Middleware](https://www.lilya.dev/middleware/) as they are both converted internally. Read more about [Python Protocols](https://peps.python.org/pep-0544/).
                """
        ),
    ] = None,
    permissions: Annotated[
        Optional[List["Permission"]],
        Doc(
            """
                A list of [permissions](https://esmerald.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
        ),
    ] = None,
    media_type: Annotated[
        Union[MediaType, str],
        Doc(
            """
                The default `media-type` used by the handler.
                """
        ),
    ] = MediaType.JSON,
    response_class: Annotated[
        Optional["ResponseType"],
        Doc(
            """
                Response class to be used within the
                handler application.

                Read more about the [Responses](https://esmerald.dev/responses/) and how
                to use them.
                """
        ),
    ] = None,
    response_cookies: Annotated[
        Optional["ResponseCookies"],
        Doc(
            """
                A sequence of `esmerald.datastructures.Cookie` objects.

                Read more about the [Cookies](https://esmerald.dev/extras/cookie-fields/?h=responsecook#cookie-from-response-cookies).
                """
        ),
    ] = None,
    response_headers: Annotated[
        Optional["ResponseHeaders"],
        Doc(
            """
                A mapping of `esmerald.datastructures.ResponseHeader` objects.

                Read more about the [ResponseHeader](https://esmerald.dev/extras/header-fields/#response-headers).
                """
        ),
    ] = None,
    tags: Annotated[
        Optional[Sequence[str]],
        Doc(
            """
                A list of strings tags to be applied to the *path operation*.

                It will be added to the generated OpenAPI documentation.

                **Note** almost everything in Esmerald can be done in [levels](https://esmerald.dev/application/levels/), which means
                these tags on a Esmerald instance, means it will be added to every route even
                if those routes also contain tags.
                """
        ),
    ] = None,
    deprecated: Annotated[
        Optional[bool],
        Doc(
            """
                Boolean flag indicating if the handler
                should be deprecated in the OpenAPI documentation.
                """
        ),
    ] = None,
    security: Annotated[
        Optional[List["SecurityScheme"]],
        Doc(
            """
                Used by OpenAPI definition, the security must be compliant with the norms.
                Esmerald offers some out of the box solutions where this is implemented.

                The [Esmerald security](https://esmerald.dev/openapi/) is available to automatically used.

                The security can be applied also on a [level basis](https://esmerald.dev/application/levels/).

                For custom security objects, you **must** subclass
                `esmerald.openapi.security.base.HTTPBase` object.
                """
        ),
    ] = None,
    operation_id: Annotated[
        Optional[str],
        Doc(
            """
                The unique identifier of the `handler`. This acts as a unique ID
                for the OpenAPI documentation.

                !!! Tip
                    Usually you don't need this as Esmerald handles it automatically
                    but it is here if you want to add your own.
                """
        ),
    ] = None,
    response_description: Annotated[
        Optional[str],
        Doc(
            """
                A description of the response. This is used for OpenAPI documentation
                purposes only and accepts all the docstrings including `markdown` format.
                """
        ),
    ] = SUCCESSFUL_RESPONSE,
    responses: Annotated[
        Optional[Dict[int, OpenAPIResponse]],
        Doc(
            """
                Additional responses that are handled by the handler and need to be described
                in the OpenAPI documentation.

                The `responses` is a dictionary like object where the first parameter is an
                `integer` and the second is an instance of an [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object.


                Read more about [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object and how to use it.
                """
        ),
    ] = None,
) -> HTTPHandler:
    """
    Handler responsible for the HTTP method `head` and
    all of its operatations.
    """

    def wrapper(func: Any) -> HTTPHandler:
        handler = HTTPHandler(
            path=path,
            methods=[HttpMethod.HEAD],
            summary=summary,
            description=description,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            response_description=response_description,
            responses=responses,
        )
        handler.fn = func
        handler.handler = func
        handler.__type__ = HttpMethod.HEAD.value
        handler.validate_handler()
        return handler

    return cast(HTTPHandler, wrapper)


def post(
    path: Annotated[
        Optional[str],
        Doc(
            """
                Relative path of the `handler`.
                The path can contain parameters in a dictionary like format
                and if the path is not provided, it will default to `/`.

                **Example**

                ```python
                @post()
                ```

                **Example with parameters**

                ```python
                @post(path="/{age: int}")
                ```
                """
        ),
    ] = None,
    *,
    summary: Annotated[
        Optional[str],
        Doc(
            """
                The summary of the handler. This short summary is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.

                **Example**

                ```python
                from esmerald import post
                from pydantic import BaseModel


                class Pretender(BaseModel):
                    name: str


                @post(summary="Black Window joining Pretenders")
                async def create_joiners(data: Pretender) -> None:
                    ...
                ```
                """
        ),
    ] = None,
    description: Annotated[
        Optional[str],
        Doc(
            """
                The description of the Esmerald application/API. This description is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.

                **Example**

                ```python
                from esmerald import post


                @post(description=...)
                async def create_joiners() -> None:
                    ...
                """
        ),
    ] = None,
    status_code: Annotated[
        Optional[int],
        Doc(
            """
                An integer indicating the status code of the handler.

                This can be achieved by passing directly the value or
                by using the `esmerald.status` or even the `lilya.status`.
                """
        ),
    ] = status.HTTP_201_CREATED,
    content_encoding: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content encoding of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    content_media_type: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content media type of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    include_in_schema: Annotated[
        bool,
        Doc(
            """
                Boolean flag indicating if it should be added to the OpenAPI docs.
                """
        ),
    ] = True,
    background: Annotated[
        Optional["BackgroundTaskType"],
        Doc(
            """
                An instance of an `esmerald.background.BackgroundTask` or `esmerald.background.BackgroundTasks` to be passed onto the handler.

                Read more about [Background tasks](https://esmerald.dev/background-tasks/) to
                understand what can be done.
                """
        ),
    ] = None,
    dependencies: Annotated[
        Optional["Dependencies"],
        Doc(
            """
                A dictionary of string and [Inject](https://esmerald.dev/dependencies/) instances enable application level dependency injection.
                """
        ),
    ] = None,
    exception_handlers: Annotated[
        Optional["ExceptionHandlerMap"],
        Doc(
            """
                A dictionary of [exception types](https://esmerald.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
        ),
    ] = None,
    middleware: Annotated[
        Optional[List["Middleware"]],
        Doc(
            """
                A list of middleware to run for every request. The middlewares of an Include will be checked from top-down or [Lilya Middleware](https://www.lilya.dev/middleware/) as they are both converted internally. Read more about [Python Protocols](https://peps.python.org/pep-0544/).
                """
        ),
    ] = None,
    permissions: Annotated[
        Optional[List["Permission"]],
        Doc(
            """
                A list of [permissions](https://esmerald.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
        ),
    ] = None,
    media_type: Annotated[
        Union[MediaType, str],
        Doc(
            """
                The default `media-type` used by the handler.
                """
        ),
    ] = MediaType.JSON,
    response_class: Annotated[
        Optional["ResponseType"],
        Doc(
            """
                Response class to be used within the
                handler application.

                Read more about the [Responses](https://esmerald.dev/responses/) and how
                to use them.

                **Example**

                ```python
                from esmerald import post, JSONResponse

                @post(response_class=JSONResponse)
                ```
                """
        ),
    ] = None,
    response_cookies: Annotated[
        Optional["ResponseCookies"],
        Doc(
            """
                A sequence of `esmerald.datastructures.Cookie` objects.

                Read more about the [Cookies](https://esmerald.dev/extras/cookie-fields/?h=responsecook#cookie-from-response-cookies).

                **Example**

                ```python
                from esmerald import post
                from esmerald.datastructures import Cookie

                response_cookies=[
                    Cookie(
                        key="csrf",
                        value="CIwNZNlR4XbisJF39I8yWnWX9wX4WFoz",
                        max_age=3000,
                        httponly=True,
                    )
                ]

                @post(response_cookies=response_cookies)
                ```
                """
        ),
    ] = None,
    response_headers: Annotated[
        Optional["ResponseHeaders"],
        Doc(
            """
                A mapping of `esmerald.datastructures.ResponseHeader` objects.

                Read more about the [ResponseHeader](https://esmerald.dev/extras/header-fields/#response-headers).

                **Example**

                ```python
                from esmerald import post
                from esmerald.datastructures import ResponseHeader

                response_headers={
                    "authorize": ResponseHeader(value="granted")
                }

                @post(response_headers=response_headers)
                ```
                """
        ),
    ] = None,
    tags: Annotated[
        Optional[Sequence[str]],
        Doc(
            """
                A list of strings tags to be applied to the *path operation*.

                It will be added to the generated OpenAPI documentation.

                **Note** almost everything in Esmerald can be done in [levels](https://esmerald.dev/application/levels/), which means
                these tags on a Esmerald instance, means it will be added to every route even
                if those routes also contain tags.

                **Example**

                ```python
                from esmerald import post

                @post(tags=["application"])
                ```
                """
        ),
    ] = None,
    deprecated: Annotated[
        Optional[bool],
        Doc(
            """
                Boolean flag indicating if the handler
                should be deprecated in the OpenAPI documentation.

                **Example**

                ```python
                from esmerald import post

                @post(deprecated=True)
                ```
                """
        ),
    ] = None,
    security: Annotated[
        Optional[List["SecurityScheme"]],
        Doc(
            """
                Used by OpenAPI definition, the security must be compliant with the norms.
                Esmerald offers some out of the box solutions where this is implemented.

                The [Esmerald security](https://esmerald.dev/openapi/) is available to automatically used.

                The security can be applied also on a [level basis](https://esmerald.dev/application/levels/).

                For custom security objects, you **must** subclass
                `esmerald.openapi.security.base.HTTPBase` object.

                **Example**

                ```python
                from esmerald import post
                from esmerald.openapi.security.http import Bearer

                @post(security=[Bearer()])
                ```
                """
        ),
    ] = None,
    operation_id: Annotated[
        Optional[str],
        Doc(
            """
                The unique identifier of the `handler`. This acts as a unique ID
                for the OpenAPI documentation.

                !!! Tip
                    Usually you don't need this as Esmerald handles it automatically
                    but it is here if you want to add your own.
                """
        ),
    ] = None,
    response_description: Annotated[
        Optional[str],
        Doc(
            """
                A description of the response. This is used for OpenAPI documentation
                purposes only and accepts all the docstrings including `markdown` format.
                """
        ),
    ] = SUCCESSFUL_RESPONSE,
    responses: Annotated[
        Optional[Dict[int, OpenAPIResponse]],
        Doc(
            """
                Additional responses that are handled by the handler and need to be described
                in the OpenAPI documentation.

                The `responses` is a dictionary like object where the first parameter is an
                `integer` and the second is an instance of an [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object.


                Read more about [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object and how to use it.


                **Example**

                ```python
                from esmerald import post
                from esmerald.openapi.datastructures import OpenAPIResponse
                from pydantic import BaseModel

                class Power(BaseModel):
                    name: str
                    description: str


                class Error(BaseModel):
                    detail: str


                @post(path='/read', responses={
                        200: OpenAPIResponse(model=Power, description=...)
                        400: OpenAPIResponse(model=Error, description=...)
                    }
                )
                async def create() -> Union[None, ItemOut]:
                    ...
                ```
                """
        ),
    ] = None,
) -> HTTPHandler:
    """
    Handler responsible for the HTTP method `post` and
    all of its operatations.

    **Example**

    ```python
    from esmerald import post


    @post()
    async def create() -> None:
        ...
    ```
    """

    def wrapper(func: Any) -> HTTPHandler:
        handler = HTTPHandler(
            path=path,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            summary=summary,
            description=description,
            methods=[HttpMethod.POST],
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            response_description=response_description,
            responses=responses,
        )
        handler.fn = func
        handler.handler = func
        handler.__type__ = HttpMethod.POST.value
        handler.validate_handler()
        return handler

    return cast(HTTPHandler, wrapper)


def put(
    path: Annotated[
        Optional[str],
        Doc(
            """
                Relative path of the `handler`.
                The path can contain parameters in a dictionary like format
                and if the path is not provided, it will default to `/`.

                **Example**

                ```python
                @put()
                ```

                **Example with parameters**

                ```python
                @put(path="/{age: int}")
                ```
                """
        ),
    ] = None,
    *,
    summary: Annotated[
        Optional[str],
        Doc(
            """
                The summary of the handler. This short summary is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.

                **Example**

                ```python
                from esmerald import put


                @put(summary="Black Window joining Pretenders")
                async def update_joiners() -> None:
                    ...
                ```
                """
        ),
    ] = None,
    description: Annotated[
        Optional[str],
        Doc(
            """
                The description of the Esmerald application/API. This description is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.

                **Example**

                ```python
                from esmerald import put


                @put(description=...)
                async def update_joiners() -> None:
                    ...
                """
        ),
    ] = None,
    status_code: Annotated[
        Optional[int],
        Doc(
            """
                An integer indicating the status code of the handler.

                This can be achieved by passing directly the value or
                by using the `esmerald.status` or even the `lilya.status`.
                """
        ),
    ] = status.HTTP_200_OK,
    content_encoding: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content encoding of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    content_media_type: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content media type of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    include_in_schema: Annotated[
        bool,
        Doc(
            """
                Boolean flag indicating if it should be added to the OpenAPI docs.
                """
        ),
    ] = True,
    background: Annotated[
        Optional["BackgroundTaskType"],
        Doc(
            """
                An instance of an `esmerald.background.BackgroundTask` or `esmerald.background.BackgroundTasks` to be passed onto the handler.

                Read more about [Background tasks](https://esmerald.dev/background-tasks/) to
                understand what can be done.
                """
        ),
    ] = None,
    dependencies: Annotated[
        Optional["Dependencies"],
        Doc(
            """
                A dictionary of string and [Inject](https://esmerald.dev/dependencies/) instances enable application level dependency injection.
                """
        ),
    ] = None,
    exception_handlers: Annotated[
        Optional["ExceptionHandlerMap"],
        Doc(
            """
                A dictionary of [exception types](https://esmerald.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
        ),
    ] = None,
    middleware: Annotated[
        Optional[List["Middleware"]],
        Doc(
            """
                A list of middleware to run for every request. The middlewares of an Include will be checked from top-down or [Lilya Middleware](https://www.lilya.dev/middleware/) as they are both converted internally. Read more about [Python Protocols](https://peps.python.org/pep-0544/).
                """
        ),
    ] = None,
    permissions: Annotated[
        Optional[List["Permission"]],
        Doc(
            """
                A list of [permissions](https://esmerald.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
        ),
    ] = None,
    media_type: Annotated[
        Union[MediaType, str],
        Doc(
            """
                The default `media-type` used by the handler.
                """
        ),
    ] = MediaType.JSON,
    response_class: Annotated[
        Optional["ResponseType"],
        Doc(
            """
                Response class to be used within the
                handler application.

                Read more about the [Responses](https://esmerald.dev/responses/) and how
                to use them.

                **Example**

                ```python
                from esmerald import put, JSONResponse

                @put(response_class=JSONResponse)
                ```
                """
        ),
    ] = None,
    response_cookies: Annotated[
        Optional["ResponseCookies"],
        Doc(
            """
                A sequence of `esmerald.datastructures.Cookie` objects.

                Read more about the [Cookies](https://esmerald.dev/extras/cookie-fields/?h=responsecook#cookie-from-response-cookies).

                **Example**

                ```python
                from esmerald import put
                from esmerald.datastructures import Cookie

                response_cookies=[
                    Cookie(
                        key="csrf",
                        value="CIwNZNlR4XbisJF39I8yWnWX9wX4WFoz",
                        max_age=3000,
                        httponly=True,
                    )
                ]

                @put(response_cookies=response_cookies)
                ```
                """
        ),
    ] = None,
    response_headers: Annotated[
        Optional["ResponseHeaders"],
        Doc(
            """
                A mapping of `esmerald.datastructures.ResponseHeader` objects.

                Read more about the [ResponseHeader](https://esmerald.dev/extras/header-fields/#response-headers).

                **Example**

                ```python
                from esmerald import put
                from esmerald.datastructures import ResponseHeader

                response_headers={
                    "authorize": ResponseHeader(value="granted")
                }

                @put(response_headers=response_headers)
                ```
                """
        ),
    ] = None,
    tags: Annotated[
        Optional[Sequence[str]],
        Doc(
            """
                A list of strings tags to be applied to the *path operation*.

                It will be added to the generated OpenAPI documentation.

                **Note** almost everything in Esmerald can be done in [levels](https://esmerald.dev/application/levels/), which means
                these tags on a Esmerald instance, means it will be added to every route even
                if those routes also contain tags.

                **Example**

                ```python
                from esmerald import put

                @put(tags=["application"])
                ```
                """
        ),
    ] = None,
    deprecated: Annotated[
        Optional[bool],
        Doc(
            """
                Boolean flag indicating if the handler
                should be deprecated in the OpenAPI documentation.

                **Example**

                ```python
                from esmerald import put

                @put(deprecated=True)
                ```
                """
        ),
    ] = None,
    security: Annotated[
        Optional[List["SecurityScheme"]],
        Doc(
            """
                Used by OpenAPI definition, the security must be compliant with the norms.
                Esmerald offers some out of the box solutions where this is implemented.

                The [Esmerald security](https://esmerald.dev/openapi/) is available to automatically used.

                The security can be applied also on a [level basis](https://esmerald.dev/application/levels/).

                For custom security objects, you **must** subclass
                `esmerald.openapi.security.base.HTTPBase` object.

                **Example**

                ```python
                from esmerald import put
                from esmerald.openapi.security.http import Bearer

                @put(security=[Bearer()])
                ```
                """
        ),
    ] = None,
    operation_id: Annotated[
        Optional[str],
        Doc(
            """
                The unique identifier of the `handler`. This acts as a unique ID
                for the OpenAPI documentation.

                !!! Tip
                    Usually you don't need this as Esmerald handles it automatically
                    but it is here if you want to add your own.
                """
        ),
    ] = None,
    response_description: Annotated[
        Optional[str],
        Doc(
            """
                A description of the response. This is used for OpenAPI documentation
                purposes only and accepts all the docstrings including `markdown` format.
                """
        ),
    ] = SUCCESSFUL_RESPONSE,
    responses: Annotated[
        Optional[Dict[int, OpenAPIResponse]],
        Doc(
            """
                Additional responses that are handled by the handler and need to be described
                in the OpenAPI documentation.

                The `responses` is a dictionary like object where the first parameter is an
                `integer` and the second is an instance of an [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object.


                Read more about [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object and how to use it.


                **Example**

                ```python
                from esmerald import put
                from esmerald.openapi.datastructures import OpenAPIResponse
                from pydantic import BaseModel

                class Power(BaseModel):
                    name: str
                    description: str


                class Error(BaseModel):
                    detail: str


                @put(path='/read', responses={
                        200: OpenAPIResponse(model=Power, description=...)
                        400: OpenAPIResponse(model=Error, description=...)
                    }
                )
                async def create() -> Union[None, ItemOut]:
                    ...
                ```
                """
        ),
    ] = None,
) -> HTTPHandler:
    """
    Handler responsible for the HTTP method `put` and
    all of its operatations.

    **Example**

    ```python
    from esmerald import put


    @put()
    async def update() -> str:
        return "updated!"
    ```
    """

    def wrapper(func: Any) -> HTTPHandler:
        handler = HTTPHandler(
            path=path,
            methods=[HttpMethod.PUT],
            summary=summary,
            description=description,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            response_description=response_description,
            responses=responses,
        )
        handler.fn = func
        handler.handler = func
        handler.__type__ = HttpMethod.PUT.value
        handler.validate_handler()
        return handler

    return cast(HTTPHandler, wrapper)


def patch(
    path: Annotated[
        Optional[str],
        Doc(
            """
                Relative path of the `handler`.
                The path can contain parameters in a dictionary like format
                and if the path is not provided, it will default to `/`.

                **Example**

                ```python
                @patch()
                ```

                **Example with parameters**

                ```python
                @patch(path="/{age: int}")
                ```
                """
        ),
    ] = None,
    *,
    summary: Annotated[
        Optional[str],
        Doc(
            """
                The summary of the handler. This short summary is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.

                **Example**

                ```python
                from esmerald import patch


                @patch(summary="Black Window joining Pretenders")
                async def update_partial_joiners() -> None:
                    ...
                ```
                """
        ),
    ] = None,
    description: Annotated[
        Optional[str],
        Doc(
            """
                The description of the Esmerald application/API. This description is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.

                **Example**

                ```python
                from esmerald import patch


                @patch(description=...)
                async def update_partial_joiners() -> None:
                    ...
                """
        ),
    ] = None,
    status_code: Annotated[
        Optional[int],
        Doc(
            """
                An integer indicating the status code of the handler.

                This can be achieved by passing directly the value or
                by using the `esmerald.status` or even the `lilya.status`.
                """
        ),
    ] = status.HTTP_200_OK,
    content_encoding: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content encoding of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    content_media_type: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content media type of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    include_in_schema: Annotated[
        bool,
        Doc(
            """
                Boolean flag indicating if it should be added to the OpenAPI docs.
                """
        ),
    ] = True,
    background: Annotated[
        Optional["BackgroundTaskType"],
        Doc(
            """
                An instance of an `esmerald.background.BackgroundTask` or `esmerald.background.BackgroundTasks` to be passed onto the handler.

                Read more about [Background tasks](https://esmerald.dev/background-tasks/) to
                understand what can be done.
                """
        ),
    ] = None,
    dependencies: Annotated[
        Optional["Dependencies"],
        Doc(
            """
                A dictionary of string and [Inject](https://esmerald.dev/dependencies/) instances enable application level dependency injection.
                """
        ),
    ] = None,
    exception_handlers: Annotated[
        Optional["ExceptionHandlerMap"],
        Doc(
            """
                A dictionary of [exception types](https://esmerald.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
        ),
    ] = None,
    middleware: Annotated[
        Optional[List["Middleware"]],
        Doc(
            """
                A list of middleware to run for every request. The middlewares of an Include will be checked from top-down or [Lilya Middleware](https://www.lilya.dev/middleware/) as they are both converted internally. Read more about [Python Protocols](https://peps.python.org/pep-0544/).
                """
        ),
    ] = None,
    permissions: Annotated[
        Optional[List["Permission"]],
        Doc(
            """
                A list of [permissions](https://esmerald.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
        ),
    ] = None,
    media_type: Annotated[
        Union[MediaType, str],
        Doc(
            """
                The default `media-type` used by the handler.
                """
        ),
    ] = MediaType.JSON,
    response_class: Annotated[
        Optional["ResponseType"],
        Doc(
            """
                Response class to be used within the
                handler application.

                Read more about the [Responses](https://esmerald.dev/responses/) and how
                to use them.

                **Example**

                ```python
                from esmerald import patch, JSONResponse

                @patch(response_class=JSONResponse)
                ```
                """
        ),
    ] = None,
    response_cookies: Annotated[
        Optional["ResponseCookies"],
        Doc(
            """
                A sequence of `esmerald.datastructures.Cookie` objects.

                Read more about the [Cookies](https://esmerald.dev/extras/cookie-fields/?h=responsecook#cookie-from-response-cookies).

                **Example**

                ```python
                from esmerald import patch
                from esmerald.datastructures import Cookie

                response_cookies=[
                    Cookie(
                        key="csrf",
                        value="CIwNZNlR4XbisJF39I8yWnWX9wX4WFoz",
                        max_age=3000,
                        httponly=True,
                    )
                ]

                @patch(response_cookies=response_cookies)
                ```
                """
        ),
    ] = None,
    response_headers: Annotated[
        Optional["ResponseHeaders"],
        Doc(
            """
                A mapping of `esmerald.datastructures.ResponseHeader` objects.

                Read more about the [ResponseHeader](https://esmerald.dev/extras/header-fields/#response-headers).

                **Example**

                ```python
                from esmerald import patch
                from esmerald.datastructures import ResponseHeader

                response_headers={
                    "authorize": ResponseHeader(value="granted")
                }

                @patch(response_headers=response_headers)
                ```
                """
        ),
    ] = None,
    tags: Annotated[
        Optional[Sequence[str]],
        Doc(
            """
                A list of strings tags to be applied to the *path operation*.

                It will be added to the generated OpenAPI documentation.

                **Note** almost everything in Esmerald can be done in [levels](https://esmerald.dev/application/levels/), which means
                these tags on a Esmerald instance, means it will be added to every route even
                if those routes also contain tags.

                **Example**

                ```python
                from esmerald import patch

                @patch(tags=["application"])
                ```
                """
        ),
    ] = None,
    deprecated: Annotated[
        Optional[bool],
        Doc(
            """
                Boolean flag indicating if the handler
                should be deprecated in the OpenAPI documentation.

                **Example**

                ```python
                from esmerald import patch

                @patch(deprecated=True)
                ```
                """
        ),
    ] = None,
    security: Annotated[
        Optional[List["SecurityScheme"]],
        Doc(
            """
                Used by OpenAPI definition, the security must be compliant with the norms.
                Esmerald offers some out of the box solutions where this is implemented.

                The [Esmerald security](https://esmerald.dev/openapi/) is available to automatically used.

                The security can be applied also on a [level basis](https://esmerald.dev/application/levels/).

                For custom security objects, you **must** subclass
                `esmerald.openapi.security.base.HTTPBase` object.

                **Example**

                ```python
                from esmerald import patch
                from esmerald.openapi.security.http import Bearer

                @patch(security=[Bearer()])
                ```
                """
        ),
    ] = None,
    operation_id: Annotated[
        Optional[str],
        Doc(
            """
                The unique identifier of the `handler`. This acts as a unique ID
                for the OpenAPI documentation.

                !!! Tip
                    Usually you don't need this as Esmerald handles it automatically
                    but it is here if you want to add your own.
                """
        ),
    ] = None,
    response_description: Annotated[
        Optional[str],
        Doc(
            """
                A description of the response. This is used for OpenAPI documentation
                purposes only and accepts all the docstrings including `markdown` format.
                """
        ),
    ] = SUCCESSFUL_RESPONSE,
    responses: Annotated[
        Optional[Dict[int, OpenAPIResponse]],
        Doc(
            """
                Additional responses that are handled by the handler and need to be described
                in the OpenAPI documentation.

                The `responses` is a dictionary like object where the first parameter is an
                `integer` and the second is an instance of an [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object.


                Read more about [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object and how to use it.


                **Example**

                ```python
                from esmerald import patch
                from esmerald.openapi.datastructures import OpenAPIResponse
                from pydantic import BaseModel

                class Power(BaseModel):
                    name: str
                    description: str


                class Error(BaseModel):
                    detail: str


                @patch(path='/read', responses={
                        200: OpenAPIResponse(model=Power, description=...)
                        400: OpenAPIResponse(model=Error, description=...)
                    }
                )
                async def update() -> Union[None, ItemOut]:
                    ...
                ```
                """
        ),
    ] = None,
) -> HTTPHandler:
    """
    Handler responsible for the HTTP method `path` and
    all of its operatations.

    **Example**

    ```python
    from esmerald import patch


    @patch()
    async def update_partial() -> str:
        return "done!"
    ```
    """

    def wrapper(func: Any) -> HTTPHandler:
        handler = HTTPHandler(
            path=path,
            methods=[HttpMethod.PATCH],
            summary=summary,
            description=description,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            response_description=response_description,
            responses=responses,
        )
        handler.fn = func
        handler.handler = func
        handler.__type__ = HttpMethod.PATCH.value
        handler.validate_handler()
        return handler

    return cast(HTTPHandler, wrapper)


def delete(
    path: Annotated[
        Optional[str],
        Doc(
            """
                Relative path of the `handler`.
                The path can contain parameters in a dictionary like format
                and if the path is not provided, it will default to `/`.

                **Example**

                ```python
                @delete()
                ```

                **Example with parameters**

                ```python
                @delete(path="/{age: int}")
                ```
                """
        ),
    ] = None,
    *,
    summary: Annotated[
        Optional[str],
        Doc(
            """
                The summary of the handler. This short summary is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.

                **Example**

                ```python
                from esmerald import delete


                @delete(summary="Black Window joining Pretenders")
                async def delete_joiners() -> None:
                    ...
                ```
                """
        ),
    ] = None,
    description: Annotated[
        Optional[str],
        Doc(
            """
                The description of the Esmerald application/API. This description is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.

                **Example**

                ```python
                from esmerald import delete


                @delete(description=...)
                async def delete_joiners() -> None:
                    ...
                """
        ),
    ] = None,
    status_code: Annotated[
        Optional[int],
        Doc(
            """
                An integer indicating the status code of the handler.

                This can be achieved by passing directly the value or
                by using the `esmerald.status` or even the `lilya.status`.
                """
        ),
    ] = status.HTTP_204_NO_CONTENT,
    content_encoding: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content encoding of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    content_media_type: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content media type of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    include_in_schema: Annotated[
        bool,
        Doc(
            """
                Boolean flag indicating if it should be added to the OpenAPI docs.
                """
        ),
    ] = True,
    background: Annotated[
        Optional["BackgroundTaskType"],
        Doc(
            """
                An instance of an `esmerald.background.BackgroundTask` or `esmerald.background.BackgroundTasks` to be passed onto the handler.

                Read more about [Background tasks](https://esmerald.dev/background-tasks/) to
                understand what can be done.
                """
        ),
    ] = None,
    dependencies: Annotated[
        Optional["Dependencies"],
        Doc(
            """
                A dictionary of string and [Inject](https://esmerald.dev/dependencies/) instances enable application level dependency injection.
                """
        ),
    ] = None,
    exception_handlers: Annotated[
        Optional["ExceptionHandlerMap"],
        Doc(
            """
                A dictionary of [exception types](https://esmerald.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
        ),
    ] = None,
    middleware: Annotated[
        Optional[List["Middleware"]],
        Doc(
            """
                A list of middleware to run for every request. The middlewares of an Include will be checked from top-down or [Lilya Middleware](https://www.lilya.dev/middleware/) as they are both converted internally. Read more about [Python Protocols](https://peps.python.org/pep-0544/).
                """
        ),
    ] = None,
    permissions: Annotated[
        Optional[List["Permission"]],
        Doc(
            """
                A list of [permissions](https://esmerald.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
        ),
    ] = None,
    media_type: Annotated[
        Union[MediaType, str],
        Doc(
            """
                The default `media-type` used by the handler.
                """
        ),
    ] = MediaType.JSON,
    response_class: Annotated[
        Optional["ResponseType"],
        Doc(
            """
                Response class to be used within the
                handler application.

                Read more about the [Responses](https://esmerald.dev/responses/) and how
                to use them.

                **Example**

                ```python
                from esmerald import delete, JSONResponse

                @delete(response_class=JSONResponse)
                ```
                """
        ),
    ] = None,
    response_cookies: Annotated[
        Optional["ResponseCookies"],
        Doc(
            """
                A sequence of `esmerald.datastructures.Cookie` objects.

                Read more about the [Cookies](https://esmerald.dev/extras/cookie-fields/?h=responsecook#cookie-from-response-cookies).

                **Example**

                ```python
                from esmerald import delete
                from esmerald.datastructures import Cookie

                response_cookies=[
                    Cookie(
                        key="csrf",
                        value="CIwNZNlR4XbisJF39I8yWnWX9wX4WFoz",
                        max_age=3000,
                        httponly=True,
                    )
                ]

                @delete(response_cookies=response_cookies)
                ```
                """
        ),
    ] = None,
    response_headers: Annotated[
        Optional["ResponseHeaders"],
        Doc(
            """
                A mapping of `esmerald.datastructures.ResponseHeader` objects.

                Read more about the [ResponseHeader](https://esmerald.dev/extras/header-fields/#response-headers).

                **Example**

                ```python
                from esmerald import delete
                from esmerald.datastructures import ResponseHeader

                response_headers={
                    "authorize": ResponseHeader(value="granted")
                }

                @delete(response_headers=response_headers)
                ```
                """
        ),
    ] = None,
    tags: Annotated[
        Optional[Sequence[str]],
        Doc(
            """
                A list of strings tags to be applied to the *path operation*.

                It will be added to the generated OpenAPI documentation.

                **Note** almost everything in Esmerald can be done in [levels](https://esmerald.dev/application/levels/), which means
                these tags on a Esmerald instance, means it will be added to every route even
                if those routes also contain tags.

                **Example**

                ```python
                from esmerald import delete

                @delete(tags=["application"])
                ```
                """
        ),
    ] = None,
    deprecated: Annotated[
        Optional[bool],
        Doc(
            """
                Boolean flag indicating if the handler
                should be deprecated in the OpenAPI documentation.

                **Example**

                ```python
                from esmerald import delete

                @delete(deprecated=True)
                ```
                """
        ),
    ] = None,
    security: Annotated[
        Optional[List["SecurityScheme"]],
        Doc(
            """
                Used by OpenAPI definition, the security must be compliant with the norms.
                Esmerald offers some out of the box solutions where this is implemented.

                The [Esmerald security](https://esmerald.dev/openapi/) is available to automatically used.

                The security can be applied also on a [level basis](https://esmerald.dev/application/levels/).

                For custom security objects, you **must** subclass
                `esmerald.openapi.security.base.HTTPBase` object.

                **Example**

                ```python
                from esmerald import delete
                from esmerald.openapi.security.http import Bearer

                @delete(security=[Bearer()])
                ```
                """
        ),
    ] = None,
    operation_id: Annotated[
        Optional[str],
        Doc(
            """
                The unique identifier of the `handler`. This acts as a unique ID
                for the OpenAPI documentation.

                !!! Tip
                    Usually you don't need this as Esmerald handles it automatically
                    but it is here if you want to add your own.
                """
        ),
    ] = None,
    response_description: Annotated[
        Optional[str],
        Doc(
            """
                A description of the response. This is used for OpenAPI documentation
                purposes only and accepts all the docstrings including `markdown` format.
                """
        ),
    ] = SUCCESSFUL_RESPONSE,
    responses: Annotated[
        Optional[Dict[int, OpenAPIResponse]],
        Doc(
            """
                Additional responses that are handled by the handler and need to be described
                in the OpenAPI documentation.

                The `responses` is a dictionary like object where the first parameter is an
                `integer` and the second is an instance of an [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object.


                Read more about [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object and how to use it.


                **Example**

                ```python
                from esmerald import delete
                from esmerald.openapi.datastructures import OpenAPIResponse
                from pydantic import BaseModel

                class Power(BaseModel):
                    name: str
                    description: str


                class Error(BaseModel):
                    detail: str


                @delete(path='/read', responses={
                        400: OpenAPIResponse(model=Error, description=...)
                        401: OpenAPIResponse(model=Power, description=...)
                    }
                )
                async def remove() -> Union[None, ItemOut]:
                    ...
                ```
                """
        ),
    ] = None,
) -> HTTPHandler:
    """
    Handler responsible for the HTTP method `delete` and
    all of its operatations.

    **Example**

    ```python
    from esmerald import delete


    @delete()
    async def remove() -> None:
        ...
    ```
    """

    def wrapper(func: Any) -> HTTPHandler:
        handler = HTTPHandler(
            path=path,
            methods=[HttpMethod.DELETE],
            summary=summary,
            description=description,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            response_description=response_description,
            responses=responses,
        )
        handler.fn = func
        handler.handler = func
        handler.__type__ = HttpMethod.DELETE.value
        handler.validate_handler()
        return handler

    return cast(HTTPHandler, wrapper)


def options(
    path: Annotated[
        Optional[str],
        Doc(
            """
                Relative path of the `handler`.
                The path can contain parameters in a dictionary like format
                and if the path is not provided, it will default to `/`.
                """
        ),
    ] = None,
    *,
    summary: Annotated[
        Optional[str],
        Doc(
            """
                The summary of the handler. This short summary is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.
                """
        ),
    ] = None,
    description: Annotated[
        Optional[str],
        Doc(
            """
                The description of the Esmerald application/API. This description is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.
                """
        ),
    ] = None,
    status_code: Annotated[
        Optional[int],
        Doc(
            """
                An integer indicating the status code of the handler.

                This can be achieved by passing directly the value or
                by using the `esmerald.status` or even the `lilya.status`.
                """
        ),
    ] = status.HTTP_200_OK,
    content_encoding: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content encoding of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    content_media_type: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content media type of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    include_in_schema: Annotated[
        bool,
        Doc(
            """
                Boolean flag indicating if it should be added to the OpenAPI docs.
                """
        ),
    ] = True,
    background: Annotated[
        Optional["BackgroundTaskType"],
        Doc(
            """
                An instance of an `esmerald.background.BackgroundTask` or `esmerald.background.BackgroundTasks` to be passed onto the handler.

                Read more about [Background tasks](https://esmerald.dev/background-tasks/) to
                understand what can be done.
                """
        ),
    ] = None,
    dependencies: Annotated[
        Optional["Dependencies"],
        Doc(
            """
                A dictionary of string and [Inject](https://esmerald.dev/dependencies/) instances enable application level dependency injection.
                """
        ),
    ] = None,
    exception_handlers: Annotated[
        Optional["ExceptionHandlerMap"],
        Doc(
            """
                A dictionary of [exception types](https://esmerald.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
        ),
    ] = None,
    middleware: Annotated[
        Optional[List["Middleware"]],
        Doc(
            """
                A list of middleware to run for every request. The middlewares of an Include will be checked from top-down or [Lilya Middleware](https://www.lilya.dev/middleware/) as they are both converted internally. Read more about [Python Protocols](https://peps.python.org/pep-0544/).
                """
        ),
    ] = None,
    permissions: Annotated[
        Optional[List["Permission"]],
        Doc(
            """
                A list of [permissions](https://esmerald.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
        ),
    ] = None,
    media_type: Annotated[
        Union[MediaType, str],
        Doc(
            """
                The default `media-type` used by the handler.
                """
        ),
    ] = MediaType.JSON,
    response_class: Annotated[
        Optional["ResponseType"],
        Doc(
            """
                Response class to be used within the
                handler application.

                Read more about the [Responses](https://esmerald.dev/responses/) and how
                to use them.
                """
        ),
    ] = None,
    response_cookies: Annotated[
        Optional["ResponseCookies"],
        Doc(
            """
                A sequence of `esmerald.datastructures.Cookie` objects.

                Read more about the [Cookies](https://esmerald.dev/extras/cookie-fields/?h=responsecook#cookie-from-response-cookies).
                """
        ),
    ] = None,
    response_headers: Annotated[
        Optional["ResponseHeaders"],
        Doc(
            """
                A mapping of `esmerald.datastructures.ResponseHeader` objects.

                Read more about the [ResponseHeader](https://esmerald.dev/extras/header-fields/#response-headers).
                """
        ),
    ] = None,
    tags: Annotated[
        Optional[Sequence[str]],
        Doc(
            """
                A list of strings tags to be applied to the *path operation*.

                It will be added to the generated OpenAPI documentation.

                **Note** almost everything in Esmerald can be done in [levels](https://esmerald.dev/application/levels/), which means
                these tags on a Esmerald instance, means it will be added to every route even
                if those routes also contain tags.
                """
        ),
    ] = None,
    deprecated: Annotated[
        Optional[bool],
        Doc(
            """
                Boolean flag indicating if the handler
                should be deprecated in the OpenAPI documentation.
                """
        ),
    ] = None,
    security: Annotated[
        Optional[List["SecurityScheme"]],
        Doc(
            """
                Used by OpenAPI definition, the security must be compliant with the norms.
                Esmerald offers some out of the box solutions where this is implemented.

                The [Esmerald security](https://esmerald.dev/openapi/) is available to automatically used.

                The security can be applied also on a [level basis](https://esmerald.dev/application/levels/).

                For custom security objects, you **must** subclass
                `esmerald.openapi.security.base.HTTPBase` object.
                """
        ),
    ] = None,
    operation_id: Annotated[
        Optional[str],
        Doc(
            """
                The unique identifier of the `handler`. This acts as a unique ID
                for the OpenAPI documentation.

                !!! Tip
                    Usually you don't need this as Esmerald handles it automatically
                    but it is here if you want to add your own.
                """
        ),
    ] = None,
    response_description: Annotated[
        Optional[str],
        Doc(
            """
                A description of the response. This is used for OpenAPI documentation
                purposes only and accepts all the docstrings including `markdown` format.
                """
        ),
    ] = SUCCESSFUL_RESPONSE,
    responses: Annotated[
        Optional[Dict[int, OpenAPIResponse]],
        Doc(
            """
                Additional responses that are handled by the handler and need to be described
                in the OpenAPI documentation.

                The `responses` is a dictionary like object where the first parameter is an
                `integer` and the second is an instance of an [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object.


                Read more about [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object and how to use it.
                """
        ),
    ] = None,
) -> HTTPHandler:
    """
    Handler responsible for the HTTP method `options` and
    all of its operatations.
    """

    def wrapper(func: Any) -> HTTPHandler:
        handler = HTTPHandler(
            path=path,
            methods=[HttpMethod.OPTIONS],
            summary=summary,
            description=description,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            response_description=response_description,
            responses=responses,
        )
        handler.fn = func
        handler.handler = func
        handler.__type__ = HttpMethod.OPTIONS.value
        handler.validate_handler()
        return handler

    return cast(HTTPHandler, wrapper)


def trace(
    path: Annotated[
        Optional[str],
        Doc(
            """
                Relative path of the `handler`.
                The path can contain parameters in a dictionary like format
                and if the path is not provided, it will default to `/`.
                """
        ),
    ] = None,
    *,
    summary: Annotated[
        Optional[str],
        Doc(
            """
                The summary of the handler. This short summary is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.
                """
        ),
    ] = None,
    description: Annotated[
        Optional[str],
        Doc(
            """
                The description of the Esmerald application/API. This description is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.
                """
        ),
    ] = None,
    status_code: Annotated[
        Optional[int],
        Doc(
            """
                An integer indicating the status code of the handler.

                This can be achieved by passing directly the value or
                by using the `esmerald.status` or even the `lilya.status`.
                """
        ),
    ] = status.HTTP_200_OK,
    content_encoding: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content encoding of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    content_media_type: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content media type of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    include_in_schema: Annotated[
        bool,
        Doc(
            """
                Boolean flag indicating if it should be added to the OpenAPI docs.
                """
        ),
    ] = True,
    background: Annotated[
        Optional["BackgroundTaskType"],
        Doc(
            """
                An instance of an `esmerald.background.BackgroundTask` or `esmerald.background.BackgroundTasks` to be passed onto the handler.

                Read more about [Background tasks](https://esmerald.dev/background-tasks/) to
                understand what can be done.
                """
        ),
    ] = None,
    dependencies: Annotated[
        Optional["Dependencies"],
        Doc(
            """
                A dictionary of string and [Inject](https://esmerald.dev/dependencies/) instances enable application level dependency injection.
                """
        ),
    ] = None,
    exception_handlers: Annotated[
        Optional["ExceptionHandlerMap"],
        Doc(
            """
                A dictionary of [exception types](https://esmerald.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
        ),
    ] = None,
    middleware: Annotated[
        Optional[List["Middleware"]],
        Doc(
            """
                A list of middleware to run for every request. The middlewares of an Include will be checked from top-down or [Lilya Middleware](https://www.lilya.dev/middleware/) as they are both converted internally. Read more about [Python Protocols](https://peps.python.org/pep-0544/).
                """
        ),
    ] = None,
    permissions: Annotated[
        Optional[List["Permission"]],
        Doc(
            """
                A list of [permissions](https://esmerald.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
        ),
    ] = None,
    media_type: Annotated[
        Union[MediaType, str],
        Doc(
            """
                The default `media-type` used by the handler.
                """
        ),
    ] = MediaType.JSON,
    response_class: Annotated[
        Optional["ResponseType"],
        Doc(
            """
                Response class to be used within the
                handler application.

                Read more about the [Responses](https://esmerald.dev/responses/) and how
                to use them.
                """
        ),
    ] = None,
    response_cookies: Annotated[
        Optional["ResponseCookies"],
        Doc(
            """
                A sequence of `esmerald.datastructures.Cookie` objects.

                Read more about the [Cookies](https://esmerald.dev/extras/cookie-fields/?h=responsecook#cookie-from-response-cookies).
                """
        ),
    ] = None,
    response_headers: Annotated[
        Optional["ResponseHeaders"],
        Doc(
            """
                A mapping of `esmerald.datastructures.ResponseHeader` objects.

                Read more about the [ResponseHeader](https://esmerald.dev/extras/header-fields/#response-headers).
                """
        ),
    ] = None,
    tags: Annotated[
        Optional[Sequence[str]],
        Doc(
            """
                A list of strings tags to be applied to the *path operation*.

                It will be added to the generated OpenAPI documentation.

                **Note** almost everything in Esmerald can be done in [levels](https://esmerald.dev/application/levels/), which means
                these tags on a Esmerald instance, means it will be added to every route even
                if those routes also contain tags.
                """
        ),
    ] = None,
    deprecated: Annotated[
        Optional[bool],
        Doc(
            """
                Boolean flag indicating if the handler
                should be deprecated in the OpenAPI documentation.
                """
        ),
    ] = None,
    security: Annotated[
        Optional[List["SecurityScheme"]],
        Doc(
            """
                Used by OpenAPI definition, the security must be compliant with the norms.
                Esmerald offers some out of the box solutions where this is implemented.

                The [Esmerald security](https://esmerald.dev/openapi/) is available to automatically used.

                The security can be applied also on a [level basis](https://esmerald.dev/application/levels/).

                For custom security objects, you **must** subclass
                `esmerald.openapi.security.base.HTTPBase` object.
                """
        ),
    ] = None,
    operation_id: Annotated[
        Optional[str],
        Doc(
            """
                The unique identifier of the `handler`. This acts as a unique ID
                for the OpenAPI documentation.

                !!! Tip
                    Usually you don't need this as Esmerald handles it automatically
                    but it is here if you want to add your own.
                """
        ),
    ] = None,
    response_description: Annotated[
        Optional[str],
        Doc(
            """
                A description of the response. This is used for OpenAPI documentation
                purposes only and accepts all the docstrings including `markdown` format.
                """
        ),
    ] = SUCCESSFUL_RESPONSE,
    responses: Annotated[
        Optional[Dict[int, OpenAPIResponse]],
        Doc(
            """
                Additional responses that are handled by the handler and need to be described
                in the OpenAPI documentation.

                The `responses` is a dictionary like object where the first parameter is an
                `integer` and the second is an instance of an [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object.


                Read more about [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object and how to use it.
                """
        ),
    ] = None,
) -> HTTPHandler:
    """
    Handler responsible for the HTTP method `trace` and
    all of its operatations.
    ```
    """

    def wrapper(func: Any) -> HTTPHandler:
        handler = HTTPHandler(
            path=path,
            methods=[HttpMethod.TRACE],
            summary=summary,
            description=description,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            response_description=response_description,
            responses=responses,
        )
        handler.fn = func
        handler.handler = func
        handler.__type__ = HttpMethod.TRACE.value
        handler.validate_handler()
        return handler

    return cast(HTTPHandler, wrapper)


def route(
    path: Annotated[
        Optional[str],
        Doc(
            """
                Relative path of the `handler`.
                The path can contain parameters in a dictionary like format
                and if the path is not provided, it will default to `/`.

                **Example**

                ```python
                @route()
                ```

                **Example with parameters**

                ```python
                @route(path="/{age: int}")
                ```
                """
        ),
    ] = None,
    *,
    methods: Annotated[
        List[str],
        Doc(
            """
                List of strings of methods allowed by the handler.

                **Example**

                ```python
                from esmerald import route


                @route(methods=["GET", "POST", "PUT", "DELETE"])
                async def handle_stuff() -> None:
                    ...
                ```
                """
        ),
    ] = None,
    summary: Annotated[
        Optional[str],
        Doc(
            """
                The summary of the handler. This short summary is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.

                **Example**

                ```python
                from esmerald import route


                @route(summary="Black Window joining Pretenders")
                async def operate() -> None:
                    ...
                ```
                """
        ),
    ] = None,
    description: Annotated[
        Optional[str],
        Doc(
            """
                The description of the Esmerald application/API. This description is displayed when the [OpenAPI](https://esmerald.dev/openapi/) documentation is used.

                **Example**

                ```python
                from esmerald import route


                @route(description=...)
                async def operate_joiners() -> None:
                    ...
                """
        ),
    ] = None,
    status_code: Annotated[
        Optional[int],
        Doc(
            """
                An integer indicating the status code of the handler.

                This can be achieved by passing directly the value or
                by using the `esmerald.status` or even the `lilya.status`.
                """
        ),
    ] = status.HTTP_200_OK,
    content_encoding: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content encoding of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    content_media_type: Annotated[
        Optional[str],
        Doc(
            """
                The string indicating the content media type of the handler.

                This is used for the generation of the [OpenAPI](https://esmerald.dev/openapi/)
                documentation.
                """
        ),
    ] = None,
    include_in_schema: Annotated[
        bool,
        Doc(
            """
                Boolean flag indicating if it should be added to the OpenAPI docs.
                """
        ),
    ] = True,
    background: Annotated[
        Optional["BackgroundTaskType"],
        Doc(
            """
                An instance of an `esmerald.background.BackgroundTask` or `esmerald.background.BackgroundTasks` to be passed onto the handler.

                Read more about [Background tasks](https://esmerald.dev/background-tasks/) to
                understand what can be done.
                """
        ),
    ] = None,
    dependencies: Annotated[
        Optional["Dependencies"],
        Doc(
            """
                A dictionary of string and [Inject](https://esmerald.dev/dependencies/) instances enable application level dependency injection.
                """
        ),
    ] = None,
    exception_handlers: Annotated[
        Optional["ExceptionHandlerMap"],
        Doc(
            """
                A dictionary of [exception types](https://esmerald.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
        ),
    ] = None,
    middleware: Annotated[
        Optional[List["Middleware"]],
        Doc(
            """
                A list of middleware to run for every request. The middlewares of an Include will be checked from top-down or [Lilya Middleware](https://www.lilya.dev/middleware/) as they are both converted internally. Read more about [Python Protocols](https://peps.python.org/pep-0544/).
                """
        ),
    ] = None,
    permissions: Annotated[
        Optional[List["Permission"]],
        Doc(
            """
                A list of [permissions](https://esmerald.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
        ),
    ] = None,
    media_type: Annotated[
        Union[MediaType, str],
        Doc(
            """
                The default `media-type` used by the handler.
                """
        ),
    ] = MediaType.JSON,
    response_class: Annotated[
        Optional["ResponseType"],
        Doc(
            """
                Response class to be used within the
                handler application.

                Read more about the [Responses](https://esmerald.dev/responses/) and how
                to use them.

                **Example**

                ```python
                from esmerald import route, JSONResponse

                @route(response_class=JSONResponse)
                ```
                """
        ),
    ] = None,
    response_cookies: Annotated[
        Optional["ResponseCookies"],
        Doc(
            """
                A sequence of `esmerald.datastructures.Cookie` objects.

                Read more about the [Cookies](https://esmerald.dev/extras/cookie-fields/?h=responsecook#cookie-from-response-cookies).

                **Example**

                ```python
                from esmerald import route
                from esmerald.datastructures import Cookie

                response_cookies=[
                    Cookie(
                        key="csrf",
                        value="CIwNZNlR4XbisJF39I8yWnWX9wX4WFoz",
                        max_age=3000,
                        httponly=True,
                    )
                ]

                @route(response_cookies=response_cookies)
                ```
                """
        ),
    ] = None,
    response_headers: Annotated[
        Optional["ResponseHeaders"],
        Doc(
            """
                A mapping of `esmerald.datastructures.ResponseHeader` objects.

                Read more about the [ResponseHeader](https://esmerald.dev/extras/header-fields/#response-headers).

                **Example**

                ```python
                from esmerald import route
                from esmerald.datastructures import ResponseHeader

                response_headers={
                    "authorize": ResponseHeader(value="granted")
                }

                @route(response_headers=response_headers)
                ```
                """
        ),
    ] = None,
    tags: Annotated[
        Optional[Sequence[str]],
        Doc(
            """
                A list of strings tags to be applied to the *path operation*.

                It will be added to the generated OpenAPI documentation.

                **Note** almost everything in Esmerald can be done in [levels](https://esmerald.dev/application/levels/), which means
                these tags on a Esmerald instance, means it will be added to every route even
                if those routes also contain tags.

                **Example**

                ```python
                from esmerald import route

                @route(tags=["application"])
                ```
                """
        ),
    ] = None,
    deprecated: Annotated[
        Optional[bool],
        Doc(
            """
                Boolean flag indicating if the handler
                should be deprecated in the OpenAPI documentation.

                **Example**

                ```python
                from esmerald import route

                @route(deprecated=True)
                ```
                """
        ),
    ] = None,
    security: Annotated[
        Optional[List["SecurityScheme"]],
        Doc(
            """
                Used by OpenAPI definition, the security must be compliant with the norms.
                Esmerald offers some out of the box solutions where this is implemented.

                The [Esmerald security](https://esmerald.dev/openapi/) is available to automatically used.

                The security can be applied also on a [level basis](https://esmerald.dev/application/levels/).

                For custom security objects, you **must** subclass
                `esmerald.openapi.security.base.HTTPBase` object.

                **Example**

                ```python
                from esmerald import route
                from esmerald.openapi.security.http import Bearer

                @route(security=[Bearer()])
                ```
                """
        ),
    ] = None,
    operation_id: Annotated[
        Optional[str],
        Doc(
            """
                The unique identifier of the `handler`. This acts as a unique ID
                for the OpenAPI documentation.

                !!! Tip
                    Usually you don't need this as Esmerald handles it automatically
                    but it is here if you want to add your own.
                """
        ),
    ] = None,
    response_description: Annotated[
        Optional[str],
        Doc(
            """
                A description of the response. This is used for OpenAPI documentation
                purposes only and accepts all the docstrings including `markdown` format.
                """
        ),
    ] = SUCCESSFUL_RESPONSE,
    responses: Annotated[
        Optional[Dict[int, OpenAPIResponse]],
        Doc(
            """
                Additional responses that are handled by the handler and need to be described
                in the OpenAPI documentation.

                The `responses` is a dictionary like object where the first parameter is an
                `integer` and the second is an instance of an [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object.


                Read more about [OpenAPIResponse](https://esmerald.dev/responses/#openapi-responses) object and how to use it.


                **Example**

                ```python
                from esmerald import route
                from esmerald.openapi.datastructures import OpenAPIResponse
                from pydantic import BaseModel

                class Power(BaseModel):
                    name: str
                    description: str


                class Error(BaseModel):
                    detail: str


                @route(path='/read', responses={
                        200: OpenAPIResponse(model=Power, description=...)
                        400: OpenAPIResponse(model=Error, description=...)
                    }
                )
                async def operate() -> Union[None, ItemOut]:
                    ...
                ```
                """
        ),
    ] = None,
) -> HTTPHandler:
    """
    Handler responsible for allowing multiple HTTP verbs in one go
    all of its operatations.

    **Example**

    ```python
    from esmerald import route


    @route(methods=["GET", "POST", "DELETE"])
    async def operate() -> str:
        return "Hello, World!"
    ```
    """
    if not methods or not isinstance(methods, list):
        raise ImproperlyConfigured(
            "http handler demands `methods` to be declared. "
            "An example would be: @route(methods=['GET', 'PUT'])."
        )

    for method in methods:
        if method.upper() not in AVAILABLE_METHODS:
            raise ImproperlyConfigured(
                f"Invalid method {method}. " "An example would be: @route(methods=['GET', 'PUT'])."
            )

    methods = [method.upper() for method in methods]
    if not status_code:  # pragma: no cover
        status_code = status.HTTP_200_OK

    def wrapper(func: Any) -> HTTPHandler:
        handler = HTTPHandler(
            path=path,
            methods=methods,
            summary=summary,
            description=description,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            response_description=response_description,
            responses=responses,
        )

        handler.fn = func
        handler.handler = func
        handler.__type__ = HttpMethod.OPTIONS.value
        handler.validate_handler()
        return handler

    return cast(HTTPHandler, wrapper)


def websocket(
    path: Annotated[
        Optional[str],
        Doc(
            """
                Relative path of the `handler`.
                The path can contain parameters in a dictionary like format
                and if the path is not provided, it will default to `/`.

                **Example**

                ```python
                @websocket()
                ```

                **Example with parameters**

                ```python
                @websocket(path="/{age: int}")
                ```
                """
        ),
    ] = None,
    *,
    dependencies: Annotated[
        Optional["Dependencies"],
        Doc(
            """
                A dictionary of string and [Inject](https://esmerald.dev/dependencies/) instances enable application level dependency injection.
                """
        ),
    ] = None,
    exception_handlers: Annotated[
        Optional["ExceptionHandlerMap"],
        Doc(
            """
                A dictionary of [exception types](https://esmerald.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
        ),
    ] = None,
    middleware: Annotated[
        Optional[List["Middleware"]],
        Doc(
            """
                A list of middleware to run for every request. The middlewares of an Include will be checked from top-down or [Lilya Middleware](https://www.lilya.dev/middleware/) as they are both converted internally. Read more about [Python Protocols](https://peps.python.org/pep-0544/).
                """
        ),
    ] = None,
    permissions: Annotated[
        Optional[List["Permission"]],
        Doc(
            """
                A list of [permissions](https://esmerald.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
        ),
    ] = None,
) -> WebSocketHandler:
    def wrapper(func: Any) -> WebSocketHandler:
        handler = WebSocketHandler(
            path=path,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
        )
        handler.fn = func
        handler.handler = func
        handler.validate_websocket_handler_function()
        return handler

    return cast(WebSocketHandler, wrapper)



from decimal import Decimal
from typing import Any, List, Type, TypeVar, Union, _GenericAlias, cast, get_args

from pydantic.fields import FieldInfo
from pydantic.json_schema import SkipJsonSchema

T = TypeVar("T", int, float, Decimal)


def is_field_optional(field: "FieldInfo") -> bool:
    """
    Returns bool True or False for the optional model field.
    """
    allow_none = getattr(field, "allow_none", True)
    return not field.is_required() and not is_any_type(field=field) and allow_none


def is_any_type(field: "FieldInfo") -> bool:
    """
    Checks if the field is of type Any.
    """
    name = cast("Any", getattr(field.annotation, "_name", None))
    return (name is not None and "Any" in name) or field.annotation is Any


def should_skip_json_schema(field_info: Union[FieldInfo, Any]) -> FieldInfo:
    """
    Checks if the schema generation for the parameters should be skipped.
    This is applied for complex fields in query params.

    Example:
        1. Union[Dict[str, str], None]
        2. Optional[Dict[str, str]]
        3. Union[List[str], None]
        4. Optional[List[str]]
    """
    union_args = get_args(field_info.annotation)

    if not any(
        isinstance(annotation, _GenericAlias) for annotation in union_args
    ) or not isinstance(field_info.annotation, _GenericAlias):
        return field_info

    arguments: List[Type[Any]] = []

    for argument in union_args:
        if argument is not type(None):
            arguments.append(argument)
        else:
            arguments.append(SkipJsonSchema[None])  # type: ignore

    arguments = tuple(arguments)  # type: ignore
    field_info.annotation = Union[arguments]
    return field_info



from typing import TYPE_CHECKING, Any, ClassVar, List, Set, Tuple, Type, cast

if TYPE_CHECKING:
    from esmerald import SimpleAPIView
    from esmerald.routing.apis.generics import ListAPIView


class SimpleAPIMeta(type):
    """
    Metaclass responsible for making sure
    only the CRUD objects are allowed.
    """

    __filtered_handlers__: ClassVar[List[str]]
    extra_allowed: List[str] = None

    def __new__(cls, name: str, bases: Tuple[Type, ...], attrs: Any) -> Any:
        """
        Making sure the `http_allowed_methods` are extended if inheritance happens
        in the subclass.

        The `http_allowed_methods` is the default for each type of generic but to allow
        extra allowed methods, the `extra_allowed` must be added.
        """
        view = super().__new__

        parents = [parent for parent in bases if isinstance(parent, SimpleAPIMeta)]
        if not parents:
            return view(cls, name, bases, attrs)

        http_allowed_methods: List[str] = []
        simple_view = cast("SimpleAPIView", view(cls, name, bases, attrs))
        cls.__filtered_handlers__ = [
            attr
            for attr in dir(simple_view)
            if not attr.startswith("__") and not attr.endswith("__")
        ]

        for base in bases:
            if (
                hasattr(base, "http_allowed_methods")
                and hasattr(base, "__is_generic__")
                and getattr(base, "__is_generic__", False) not in [False, None]
            ):
                http_allowed_methods.extend(base.http_allowed_methods)

        if hasattr(simple_view, "extra_allowed") and simple_view.extra_allowed is not None:
            assert isinstance(
                simple_view.extra_allowed, list
            ), "`extra_allowed` must be a list of strings allowed."

            http_allowed_methods.extend(simple_view.extra_allowed)

        http_allowed_methods.extend(simple_view.http_allowed_methods)

        # Remove any duplicates
        allowed_methods: Set[str] = {method.lower() for method in http_allowed_methods}

        # Reasign the new clean list
        simple_view.http_allowed_methods = list(allowed_methods)
        for handler_name in cls.__filtered_handlers__:
            for base in simple_view.__bases__:
                attribute = getattr(simple_view, handler_name)
                simple_view.is_method_allowed(handler_name, base, attribute)
        return simple_view


class ListAPIMeta(SimpleAPIMeta):
    """
    Metaclass with an extra for lists specifically.
    """

    def __new__(cls, name: str, bases: Tuple, attrs: Any) -> Any:
        view: "ListAPIView" = super().__new__(cls, name, bases, attrs)

        if not hasattr(view, "__filtered_handlers__"):
            return view

        for handler_name in view.__filtered_handlers__:
            for base in view.__bases__:
                attribute = getattr(view, handler_name)
                view.is_signature_valid(handler_name, base, attribute, signature_type=list)
        return view
